CLIS_DESC = [("environment", ".environment_cli.cli")]

FLOW_DECORATORS_DESC = [
    ("conda_base", ".conda.conda_flow_decorator.CondaRequirementFlowDecorator"),
    ("pip_base", ".conda.conda_flow_decorator.PipRequirementFlowDecorator"),
    ("pypi_base", ".conda.conda_flow_decorator.PypiRequirementFlowDecorator"),
    ("named_env_base", ".conda.conda_flow_decorator.NamedEnvRequirementFlowDecorator"),
]
STEP_DECORATORS_DESC = [
    ("conda", ".conda.conda_step_decorator.CondaRequirementStepDecorator"),
    ("pip", ".conda.conda_step_decorator.PipRequirementStepDecorator"),
    ("pypi", ".conda.conda_step_decorator.PypiRequirementStepDecorator"),
    ("named_env", ".conda.conda_step_decorator.NamedEnvRequirementStepDecorator"),
    ("conda_env_internal", ".conda.conda_step_decorator.CondaEnvInternalDecorator"),
]
ENVIRONMENTS_DESC = [("conda", ".conda.conda_environment.CondaEnvironment")]

__mf_promote_submodules__ = ["conda"]



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false
import os
import re
import tempfile

from itertools import chain
from typing import Dict, List, Optional, Set, Tuple, TYPE_CHECKING, cast  # noqa

from urllib.parse import urlparse
from metaflow.debug import debug

# Consider re-adding if resolving fails weirdly.
# from metaflow.exception import retry_exp_backoff

from metaflow._vendor.packaging.requirements import Requirement

from ..env_descr import (
    CondaPackageSpecification,
    EnvType,
    PypiCachePackage,
    PypiPackageSpecification,
    ResolvedEnvironment,
)

from ..pypi_package_builder import PackageToBuild, build_pypi_packages

from ..utils import (
    CondaException,
    arch_id,
    channel_or_url,
    filter_user_reqs_by_markers,
    get_requirement,
    parse_explicit_url_conda,
    parse_explicit_url_pypi,
    pypi_tags_from_arch,
    split_into_dict,
)
from . import Resolver
from .pip_resolver import _FAKE_WHEEL

if TYPE_CHECKING:
    from metaflow_extensions.netflix_ext.plugins.conda.env_descr import (
        PackageSpecification,
    )  # noqa


class CondaLockResolver(Resolver):
    TYPES = ["conda-lock"]

    # @retry_exp_backoff(
    #     (CondaException, OSError),
    #     "Failed to resolve conda-lock environment",
    #     tries=3,
    #     max_delay_per_retry=5,
    # )
    def resolve(
        self,
        env_type: EnvType,
        python_version_requested: str,
        deps: Dict[str, List[str]],
        sources: Dict[str, List[str]],
        extras: Dict[str, List[str]],
        architecture: str,
        builder_envs: Optional[List[ResolvedEnvironment]] = None,
        base_env: Optional[ResolvedEnvironment] = None,
        file_paths: Dict[str, List[str]] = {},
    ) -> Tuple[ResolvedEnvironment, Optional[List[ResolvedEnvironment]]]:
        outfile_name = None
        external_packages = {}
        if base_env:
            local_packages = [
                p for p in base_env.packages if not p.is_downloadable_url()
            ]
            external_packages = {
                p.package_name: p.url
                for p in base_env.packages
                if p.TYPE == "pypi" and p.is_external_url(sources)  # type: ignore[attr-defined]
            }
            if local_packages:
                # We actually only care about things that are not online. Derived packages
                # are OK because we can reconstruct them if needed (or they may even
                # be cached)

                raise CondaException(
                    "Local PYPI packages are not supported in MIXED mode: %s"
                    % ", ".join([p.package_name for p in local_packages])
                )

        # self._start_micromamba_server()

        index_url = self._conda.default_pypi_sources[0]
        # pypi_channels contains everything EXCEPT the pypi default because we don't
        # need to add that to poetry
        pypi_channels = list(
            set(sources.get("pypi", [])).difference(["https://pypi.org/simple"])
        )
        debug.conda_exec("Will add pypi channels: %s" % ", ".join(pypi_channels))

        deps = filter_user_reqs_by_markers(deps, python_version_requested, architecture)
        try:
            # We resolve the environment using conda-lock

            # Write out the TOML file. It's easy enough that we don't use another tool
            # to write it out. We use TOML so that we can disable pypi if needed

            pypi_deps = deps.get("pypi", [])
            conda_deps = list(chain(deps.get("conda", []), deps.get("npconda", [])))

            sys_overrides = split_into_dict(deps.get("sys", []))

            # We only add pip if not present
            if not any([(d == "pip" or d.startswith("pip==")) for d in conda_deps]):
                conda_deps.append("pip")
            # Ditto for uv -- uv 0.7.8 is the last to actually work with python 3.7
            # but this is not listed as a condition so we enforce it here
            # TODO: Remove this once python3.7 is retired
            if not any([(d == "uv" or d.startswith("uv==")) for d in conda_deps]):
                # Duplicate code from filter_user_reqs_by_markers in EnvsResolver.
                # Get the minimum python version supported (major.minor) from the requested specifier.
                spec = get_requirement(python_version_requested).specifier
                candidate_versions = [
                    "3.5",
                    "3.6",
                    "3.7",
                    "3.8",
                    "3.9",
                    "3.10",
                    "3.11",
                    "3.12",
                    "3.13",
                    "3.14",
                ]
                matching_versions = list(spec.filter(candidate_versions))
                if matching_versions:
                    min_supported_py = matching_versions[0]
                else:
                    # fallback to the requested version (eg, ==3.10.5 → 3.10)
                    mm = re.match(r"^.*?([0-9]+)\.([0-9]+)", python_version_requested)
                    if mm:
                        min_supported_py = f"{mm.group(1)}.{mm.group(2)}"
                    else:
                        raise CondaException(
                            "Could not determine minimum supported Python "
                            f"version from {python_version_requested}"
                        )
                if min_supported_py in ["3.5", "3.6", "3.7"]:
                    conda_deps.append("uv==<=0.7.8")
                else:
                    conda_deps.append("uv")

            toml_lines = [
                "[build-system]\n",
                'requires = ["poetry>=0.12"]\n',
                'build-backend = "poetry.masonry.api"\n',
                "\n",
                "[tool.conda-lock]\n",
            ]
            # Add channels
            have_extra_channels = (
                len(
                    set(sources.get("conda", [])).difference(
                        map(channel_or_url, self._conda.default_conda_channels)
                    )
                )
                > 0
            )

            toml_lines.append(
                "channels = [%s]\n"
                % ", ".join(["'%s'" % c for c in sources.get("conda", [])])
            )

            toml_lines.append(
                "pip-repositories = [%s]\n"
                % ", ".join(["'%s'" % c for c in pypi_channels])
            )
            if index_url != "https://pypi.org/simple":
                toml_lines.append("allow-pypi-requests = false\n")

            toml_lines.append("\n")
            # TODO: Maybe we can make this better and only relax if :: is for channels
            # that don't exist in the list
            if any(["::" in d for d in conda_deps]) or have_extra_channels:
                addl_env = {"CONDA_CHANNEL_PRIORITY": "flexible"}
            else:
                addl_env = {}

            # Add deps
            toml_lines.append("[tool.conda-lock.dependencies]\n")
            seen = set()  # type: Set[str]
            for d in conda_deps:
                splits = d.split("==", 1)
                if len(splits) == 2:
                    if splits[0] in seen:
                        raise CondaException(
                            "Duplicate conda dependency %s" % splits[0]
                        )
                    toml_lines.append('"%s" = "%s"\n' % (splits[0], splits[1]))
                    seen.add(splits[0])
                else:
                    if d in seen:
                        raise CondaException("Duplicate conda dependency %s" % d)
                    toml_lines.append('"%s" = "*"\n' % d)
                    seen.add(d)
            toml_lines.append("\n")
            toml_lines.append("[tool.poetry.dependencies]\n")
            # In some cases (when we build packages), we may actually have the same
            # dependency multiple times. We keep just the URL one in this case
            pypi_dep_lines = {}  # type: Dict[str, Dict[str, str]]
            for d in pypi_deps:
                splits = d.split("==", 1)
                # Here we re-parse the requirement. It will be one of the four options:
                #  - <package_name>
                #  - <package_name>[extras]
                #  - <package_name>@<url>
                #  - <package_name>[extras]@<url>

                # We first check if we recorded an external URL for this package --
                # if so, we need to indicate that to conda-lock
                external_url = external_packages.get(splits[0], None)
                if external_url:
                    pypi_dep_lines.setdefault(splits[0], {}).update(
                        {"url": external_url, "url_extras": ""}
                    )
                    continue

                # If not external URL, parse the requirement according to cases above
                parsed_req = Requirement(splits[0])
                if parsed_req.extras:
                    extra_part = "extras = [%s]," % ", ".join(
                        ['"%s"' % e for e in parsed_req.extras]
                    )
                else:
                    extra_part = ""

                version_str = splits[1] if len(splits) == 2 else "*"
                if parsed_req.url:
                    if len(splits) == 2:
                        raise CondaException(
                            "Unexpected version on URL requirement %s" % splits[0]
                        )
                    pypi_dep_lines.setdefault(parsed_req.name, {}).update(
                        {"url": parsed_req.url, "url_extras": extra_part}
                    )
                else:
                    pypi_dep_lines.setdefault(parsed_req.name, {}).update(
                        {"version": version_str, "extras": extra_part}
                    )
            for pypi_name, info in pypi_dep_lines.items():
                if "url" in info:
                    toml_lines.append(
                        '"%s" = {%s = "%s", %s source="pypi"}\n'
                        % (
                            pypi_name,
                            "git" if info["url"].startswith("git+") else "url",
                            info["url"],
                            info["url_extras"],
                        )
                    )
                else:
                    toml_lines.append(
                        '"%s" = {version = "%s", %s source="pypi"}\n'
                        % (pypi_name, info["version"], info["extras"])
                    )

            with tempfile.TemporaryDirectory() as conda_lock_dir:
                outfile_name = "/tmp/conda-lock-gen-%s" % os.path.basename(
                    conda_lock_dir
                )

                args = [
                    "lock",
                    "-f",
                    "pyproject.toml",
                    "-p",
                    architecture,
                    "--filename-template",
                    outfile_name,
                    "-k",
                    "explicit",
                    "--conda",
                ]
                #                if "micromamba_server" in self._conda._bins:
                #                    args.extend([self._bins["micromamba_server"], "--micromamba"])
                #                else:
                conda_exec_type = self._conda.conda_executable_type
                if conda_exec_type:
                    args.append(cast(str, self._conda.binary(conda_exec_type)))

                    if conda_exec_type != "conda":
                        args.append("--%s" % conda_exec_type)
                else:
                    raise CondaException("Could not find conda binary for conda-lock")

                # If arch_id() == architecture, we also use the same virtual packages
                # as the ones that exist on the machine to mimic the current behavior
                # of conda/mamba. If not the same architecture, we pass the system
                # overrides (currently __cuda) down as well.
                if arch_id() == architecture or sys_overrides:
                    lines = ["subdirs:\n", "  %s:\n" % architecture, "    packages:\n"]
                    # Strip out build as it seems to be causing some issues with
                    # some versions of conda-lock when it extracts the GLIBC version.
                    lines.extend(
                        '      %s: "%s"\n' % (virt_pkg, virt_build_str.split("=")[0])
                        for virt_pkg, virt_build_str in self._conda.virtual_packages.items()
                        if virt_pkg not in sys_overrides
                    )
                    lines.extend(
                        '      %s: "%s"\n' % (k, v.split("=")[0])
                        for k, v in sys_overrides.items()
                    )

                    with open(
                        os.path.join(conda_lock_dir, "virtual_yml.spec"),
                        mode="w",
                        encoding="ascii",
                    ) as virtual_yml:
                        virtual_yml.writelines(lines)
                        debug.conda_exec(
                            "Virtual package configuration:\n%s" % "".join(lines)
                        )
                    args.extend(["--virtual-package-spec", "virtual_yml.spec"])

                debug.conda_exec("Build directory: %s" % conda_lock_dir)
                # conda-lock will only consider a `pyproject.toml` as a TOML file which
                # is somewhat annoying.
                with open(
                    os.path.join(conda_lock_dir, "pyproject.toml"),
                    mode="w",
                    encoding="ascii",
                ) as input_toml:
                    input_toml.writelines(toml_lines)
                    debug.conda_exec("TOML configuration:\n%s" % "".join(toml_lines))
                self._conda.call_binary(
                    args, binary="conda-lock", addl_env=addl_env, cwd=conda_lock_dir
                )
            # At this point, we need to read the explicit dependencies in the file created
            emit = False
            packages = []  # type: List["PackageSpecification"]
            packages_to_build = {}  # type: Dict[str, PackageToBuild]
            with open(outfile_name, "r", encoding="utf-8") as out:
                for l in out:
                    if emit:
                        if l.startswith("#"):
                            components = l.split()
                            # Line should be # pip <pkg> @ <url>
                            if len(components) != 5:
                                raise CondaException(
                                    "Unexpected package specification line: %s" % l
                                )
                            debug.conda_exec("Got PYPI package %s" % l)
                            if components[4].startswith("git+"):
                                # This is a GIT URL so we have to build the package
                                # See pip_resolver.py for vcs_info on an explanation
                                # on how we deal with git packages
                                # Looks like gi+https://<url>@<commit>
                                base_build_url = components[4]
                                parse = urlparse(base_build_url)
                                clean_path, clean_commit = parse.path.split("@")
                                clean_url = parse.scheme[4:] + parse.netloc + clean_path
                                base_pkg_url = "%s/%s" % (clean_url, clean_commit)
                                # TODO: Do we need to handle subdirectories
                                cache_base_url = (
                                    PypiCachePackage.make_partial_cache_url(
                                        base_pkg_url, is_real_url=False
                                    )
                                )
                                packages_to_build[cache_base_url] = PackageToBuild(
                                    base_build_url,
                                    PypiPackageSpecification(
                                        _FAKE_WHEEL,
                                        base_pkg_url,
                                        is_real_url=False,
                                        url_format=".whl",
                                    ),
                                )
                                continue
                            # Non-GIT URL
                            parse_result = parse_explicit_url_pypi(components[4])
                            if parse_result.url_format != ".whl":
                                cache_base_url = (
                                    PypiCachePackage.make_partial_cache_url(
                                        parse_result.url, is_real_url=True
                                    )
                                )
                                packages_to_build[cache_base_url] = PackageToBuild(
                                    parse_result.url,
                                    PypiPackageSpecification(
                                        parse_result.filename,
                                        parse_result.url,
                                        is_real_url=True,
                                        url_format=parse_result.url_format,
                                    ),
                                )
                            else:
                                packages.append(
                                    PypiPackageSpecification(
                                        parse_result.filename,
                                        parse_result.url,
                                        url_format=parse_result.url_format,
                                        hashes=(
                                            {parse_result.url_format: parse_result.hash}
                                            if parse_result.hash
                                            else None
                                        ),
                                    )
                                )
                        else:
                            parse_result = parse_explicit_url_conda(l.strip())
                            packages.append(
                                CondaPackageSpecification(
                                    parse_result.filename,
                                    parse_result.url,
                                    url_format=parse_result.url_format,
                                    hashes={
                                        parse_result.url_format: cast(
                                            str, parse_result.hash
                                        )
                                    },
                                )
                            )
                    if not emit and l.strip() == "@EXPLICIT":
                        emit = True
            if packages_to_build:
                with tempfile.TemporaryDirectory() as build_dir:
                    python_version = None  # type: Optional[str]
                    for p in packages:
                        if p.filename.startswith("python-"):
                            python_version = p.package_version
                            break
                    if python_version is None:
                        raise CondaException(
                            "Could not determine version of Python from conda packages"
                        )
                    if architecture == "linux-64":
                        # Get the latest supported GLIBC version so we can generate
                        # the proper tags (only matters on linux)
                        glibc_version_candidates = [
                            d for d in deps.get("sys", []) if d.startswith("__glibc")
                        ]
                        if len(glibc_version_candidates) != 1:
                            raise CondaException(
                                "Could not determine maximum GLIBC version"
                            )
                        # Get version looking like 2.27=0
                        glibc_version = glibc_version_candidates[0][len("__glibc==") :]
                        # Strip =0
                        glibc_version = glibc_version.split("=", 1)[0]
                        # Replace . with _
                        glibc_version = glibc_version.replace(".", "_")
                    else:
                        glibc_version = ""
                    supported_tags = pypi_tags_from_arch(
                        python_version, architecture, glibc_version
                    )
                    if not self._conda.storage:
                        raise CondaException(
                            "Cannot create a relocatable environment as it depends on "
                            "local files or non wheels and no storage backend is defined: %s"
                            % ", ".join([v.url for v in packages_to_build.values()])
                        )
                    built_pypi_packages, builder_envs = build_pypi_packages(
                        self._conda,
                        self._conda.storage,
                        python_version,
                        packages_to_build,
                        builder_envs,
                        build_dir,
                        architecture,
                        supported_tags,
                        sources.get("pypi", []),
                    )
                    packages.extend(built_pypi_packages)
            return (
                ResolvedEnvironment(
                    deps,
                    sources,
                    extras,
                    architecture,
                    all_packages=packages,
                    env_type=env_type,
                ),
                builder_envs,
            )
        finally:
            if outfile_name and os.path.isfile(outfile_name):
                os.unlink(outfile_name)



_jupyter_instructions_markdown = """
### Instructions for Debugging 

##### 1. Execute the previous cells before running the code cell below.

##### 2. Select the correct kernel for the notebook.
- By default, the kernel was selected based on the environment we created for 
debugging the task. In case the kernel is not selected correctly, 
follow the steps given below.
- If the task originally ran in a conda environment, select the kernel whose 
display name starts with `Metaflow ...`. Else, select any python kernel 
available.

##### 3. Add the code for the Task in the cells below!
- The code for the task is in the file `{FileName}` and the step begins from 
line no: `{StepStartLine}` and ends at line no: `{StepEndLine}`. 
- You can copy the code inside the function definition and execute it line 
  by line (or however you want) to debug the task as you wish. 
- The function definition has also been provided below for reference:

```python
{FunctionDefinition}
```
"""



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false
from typing import Dict, List, Optional, Tuple

from ..env_descr import (
    EnvType,
    ResolvedEnvironment,
)

from . import Resolver


class CondaLockResolver(Resolver):
    TYPES = ["micromamba_server"]

    def resolve(
        self,
        env_type: EnvType,
        deps: Dict[str, List[str]],
        sources: Dict[str, List[str]],
        extras: Dict[str, List[str]],
        architecture: str,
        builder_envs: Optional[List[ResolvedEnvironment]] = None,
        base_env: Optional[ResolvedEnvironment] = None,
    ) -> Tuple[ResolvedEnvironment, Optional[List[ResolvedEnvironment]]]:
        raise NotImplementedError

    # The below code is not maintained and kept here as a reference -- it probably
    # needs to be cleaned up
    # deps = [d for d in deps if d.category in ("conda", "npconda")]

    # if not self._have_micromamba_server:
    #     raise CondaException(
    #         "Micromamba server not supported by installed version of micromamba"
    #     )

    # self._start_micromamba_server()
    # # Form the payload to send to the server
    # req = {
    #     "specs": [d.value for d in deps if d.category in ("conda", "npconda")]
    #     + ["pip"],
    #     "platform": architecture,
    #     "channels": [c.value for c in channels if c.category == "conda"],
    # }
    # if arch_id() == architecture:
    #     # Use the same virtual packages as the ones used for conda/mamba
    #     req["virtual_packages"] = [
    #         "%s=%s" % (virt_pkg, virt_build_str)
    #         for virt_pkg, virt_build_str in self.virtual_packages.items()
    #     ]
    # # Make the request to the micromamba server
    # debug.conda_exec(
    #     "Payload for micromamba server on port %d: %s"
    #     % (self._micromamba_server_port, str(req))
    # )
    # resp = requests.post(
    #     "http://localhost:%d" % self._micromamba_server_port, json=req
    # )
    # if resp.status_code != 200:
    #     raise CondaException(
    #         "Got unexpected return code from micromamba server: %d"
    #         % resp.status_code
    #     )
    # else:
    #     json_response = resp.json()
    #     if "error_msg" in json_response:
    #         raise CondaException(
    #             "Cannot resolve environment: %s" % json_response["error_msg"]
    #         )
    #     else:
    #         return [
    #             CondaPackageSpecification(
    #                 filename=pkg["filename"],
    #                 url=pkg["url"],
    #                 url_format=correct_splitext(pkg["filename"])[1],
    #                 hashes={correct_splitext(pkg["filename"])[1]: pkg["md5"]},
    #             )
    #             for pkg in json_response["packages"]
    #         ]



from ..env_descr import (
    PackageSpecification,
    PypiPackageSpecification,
    ResolvedEnvironment,
    EnvType,
)

from ..utils import (
    filter_packages_by_markers,
    get_best_compatible_packages,
    get_maximum_glibc_version,
    get_python_full_version_from_builder_envs,
    parse_explicit_url_pypi_no_hash,
    pypi_tags_from_arch,
)

from metaflow._vendor.packaging.tags import (
    Tag,
)

from metaflow_extensions.netflix_ext.plugins.conda.conda import CondaException

from . import Resolver
from typing import Callable, Dict, Iterable, List, Any, Optional, Tuple
import os
import itertools
from itertools import chain


class PylockTomlResolver(Resolver):
    TYPES = ["pylock_toml"]
    REQUIRES_BUILDER_ENV = True

    """
    Resolve the environment based on the provided pylock.toml file.

    The pylock.toml path will be passed in file_paths['pylock_toml']
    """

    def resolve(
        self,
        env_type: EnvType,
        python_version_requested: str,
        deps: Dict[str, List[str]],
        sources: Dict[str, List[str]],
        extras: Dict[str, List[str]],
        architecture: str,
        builder_envs: Optional[List[ResolvedEnvironment]] = None,
        base_env: Optional[ResolvedEnvironment] = None,
        file_paths: Dict[str, List[str]] = {},
    ) -> Tuple[ResolvedEnvironment, Optional[List[ResolvedEnvironment]]]:
        # Checks on len(tom_path_list) >= 1 skiped due to get_resolver_cls()
        # only triggers this class's resolve() call when its len >= 1
        # builder_envs check also skipped.

        toml_path_list = file_paths["pylock_toml"]

        toml_path = toml_path_list[0]
        if not (
            toml_path is not None
            and isinstance(toml_path, str)
            and toml_path.strip() != ""
        ):
            raise ValueError("pylock_toml_path must be a non-empty string")

        if not os.path.isfile(toml_path):
            raise ValueError(f"TOML file does not exist: {toml_path}")

        toml_root_obj = self._read_toml(toml_path)

        if not builder_envs:
            raise CondaException(
                "PylockTomlResolver requires a non-empty builder_envs. Please contact metaflow team."
            )
        base_conda_envs = [r for r in builder_envs if r.env_id.arch == architecture]
        if not base_conda_envs:
            raise CondaException(
                "Could not find a builder environment for architecture {architecture}. "
                + "Please contact the Metaflow team."
            )

        base_conda_env = base_conda_envs[0]

        glibc_version = get_maximum_glibc_version(architecture, deps)
        python_full_version = get_python_full_version_from_builder_envs(
            builder_envs, architecture
        )
        if python_full_version is None:
            raise CondaException(
                "Could not determine version of Python from conda packages"
            )

        supported_tags = pypi_tags_from_arch(
            python_full_version,
            architecture,
            glibc_version,
        )

        def filter_func(
            package_dict: Dict[str, List[PackageSpecification]]
        ) -> Dict[str, List[PackageSpecification]]:
            return PylockTomlResolver._filter_packages(
                package_dict,
                python_full_version,
                architecture,
            )

        resolved_env = self._translate_pylock_toml_to_resolved_env(
            toml_root_obj,
            env_type,
            deps,
            list(base_conda_env.packages),
            sources,
            extras,
            architecture,
            supported_tags,
            filter_func,
        )

        return (resolved_env, builder_envs)

    """
    At this time, we will only support the wheels format, and explicitly reject vcs, archive, directory
    formats.

    CONTEXT:

    The formats of packages in pylock.toml (https://peps.python.org/pep-0751/) are:
    vcs, archive, sdist+wheels, directory.
    1. The formats are specified by the existence of one of the four inline-tables (subtables): [packages.vcs], [packages.directory],
        [packages.sdist] within a [[package]] table. These four subtables are mutually exclusive
        within the same package. This design ensures that a package can be specified with only one format.
    2. [[packages.Wheels]] is a nested array table mutually exclusive with other formats except [packages.sdist], which makes
       it a nested array only available within a [packages.sdist].
        (More details can be found here: https://peps.python.org/pep-0751/#packages)

    RATIONALE for supporting wheels only in this version:

    1. Despite the user passions in adoption of uv, we want a gradual rollout of support for other formats,
       to ensure our thorough testing.
    2. We do not want our customers to be the first ones to encounter issues with them.
       We will officially support those only after a thorough testing, with abundant real-world
       pylock.py examples with those formats (which we don't have).
    3. We don't want to skip parsing of a particular package when it's vcs/archive/directory,
       a practice which makes the silent failures much harder to debug.
    """
    UNSUPPORTED_PACKAGE_SUBTYPES = {"vcs", "archive", "directory"}

    # TODO(https://netflix.atlassian.net/browse/MLPMF-502): wait for dependency of pylock.py to resolved, so that we could use
    # the pylock.py to read it.
    @staticmethod
    def _read_toml(path: str) -> Dict[str, Any]:
        try:
            import tomli
        except ImportError:
            raise RuntimeError(
                "Could not import the tomli library to use pylock_toml_resolver (or uv decorator). "
                + "We intentionally did not include this as a Metaflow's required dependency. Please "
                + "manually install it when you see this message."
            )

        with open(path, "rb") as f:
            return tomli.load(f)

    @staticmethod
    def _filter_packages(
        packages_dict: Dict[str, List[PackageSpecification]],
        python_full_version: str,
        architecture: str,
        target_env: Optional[Dict[str, str]] = None,
    ) -> Dict[str, List[PackageSpecification]]:

        for key, pkgs in packages_dict.items():
            for pkg in pkgs:
                if key != pkg._package_name:
                    raise CondaException(
                        "Package dict's key should equal to each package's _package_name attribute."
                        f"key={key}, pkg._package_name={pkg._package_name}. Please contact Metaflow team."
                    )

        flattened_packages = list(chain.from_iterable(packages_dict.values()))
        filtered_packages = filter_packages_by_markers(
            flattened_packages,
            python_full_version,
            architecture,
            target_env,
        )

        # TODO: check this name is equal to the previous package dict name.
        filtered_packages.sort(key=lambda pkg: pkg._package_name)
        out_dict = dict(
            (k, list(v))
            for k, v in itertools.groupby(
                filtered_packages, key=lambda pkg: pkg._package_name
            )
        )
        return out_dict

    @staticmethod
    def _translate_pylock_toml_to_resolved_env(
        toml_root_obj: Dict[str, Any],
        env_type: EnvType,
        deps: Dict[str, List[str]],
        base_packages: Iterable[PackageSpecification],
        sources: Dict[str, List[str]],
        extras: Dict[str, List[str]],
        architecture: str,
        supported_tags: List[Tag],
        # We take a Callable here rather than do the filtering directly
        # in this function due to:
        # 1. The filtering call requires knowledge of architecture, python_version,
        #    target_env, which are not directly available in this function,
        #    the bringing in of which complicates this function.
        # 2. Possibility of a centralized filtering logic for all file formats.
        # 3. Easier for unit testing and disabling of filtering.
        package_filter_func: Optional[
            Callable[
                [Dict[str, List[PackageSpecification]]],
                Dict[str, List[PackageSpecification]],
            ]
        ] = None,
    ) -> ResolvedEnvironment:

        packages_dict = PylockTomlResolver._pylock_toml_root_obj_to_packages(
            toml_root_obj
        )

        if package_filter_func:
            packages_dict = package_filter_func(packages_dict)

        best_packages = get_best_compatible_packages(
            packages_dict,
            supported_tags,
        )

        packages: List[PackageSpecification] = list(best_packages.values())
        packages.extend(base_packages)

        resolved_env = PylockTomlResolver._packages_to_resolved_env(
            packages,
            env_type,
            deps,
            sources,
            extras,
            architecture,
        )

        return resolved_env

    @staticmethod
    # TODO(https://netflix.atlassian.net/browse/MLPMF-502): use the pylock.py object instead.
    def _toml_package_to_package_specs(
        toml_package_dict: Dict[str, Any],
    ) -> List[PackageSpecification]:
        """
        Convert a TOML package item to a PackageSpecification object.

        Sample package item from pylock.toml:

        [[packages]]
        name = "certifi"
        version = "2025.8.3"
        sdist = { url = "https://pypi.netflix.net/packages/18933835518/certifi-2025.8.3.tar.gz", upload-time = 2025-08-03T03:07:47Z, size = 162386, hashes = { sha256 = "e564105f78ded564e3ae7c923924435e1daa7463faeab5bb932bc53ffae63407" } }
        wheels = [{ url = "https://pypi.netflix.net/packages/18933835517/certifi-2025.8.3-py3-none-any.whl", upload-time = 2025-08-03T03:07:45Z, size = 161216, hashes = { sha256 = "f6c12493cfb1b06ba2ff328595af9350c65d6644968e5d3a2ffd78699af217a5" } }]
        """

        if PylockTomlResolver.UNSUPPORTED_PACKAGE_SUBTYPES & set(
            toml_package_dict.keys()
        ):
            raise CondaException(
                "Currently we only support wheel packages. Please contact the Metaflow team "
                + "if you need us to support more general cases.\n"
                + "And please mention the following information to the Metaflow team:\n"
                f"toml_package_dict.keys()={toml_package_dict.keys()}"
            )

        wheels = toml_package_dict.get("wheels", [])
        if not wheels:
            raise CondaException(
                "We just encountered a package definition that contains no wheels.\n"
                + "Please help us diagnose the issue by mentioning the following information to the Metaflow team:\n"
                f"toml_package_dict={toml_package_dict}"
            )

        packages: List[PackageSpecification] = []

        # This set of attributes will be applied to all wheels within this package.
        package_attributes = {
            k: toml_package_dict.get(k)
            for k in ["name", "version", "marker"]
            if k in toml_package_dict
        }

        for wheel in wheels:
            url = wheel.get("url", "")
            file_hashes = wheel.get("hashes", {})

            if url and file_hashes:
                parse_result = parse_explicit_url_pypi_no_hash(url)
                package = PypiPackageSpecification(
                    filename=parse_result.filename,
                    url=url,
                    hashes=file_hashes,
                    # Confirmed marker will only be a string in pylock.toml
                    # Thread: https://netflix.slack.com/archives/C0D3Z78F4/p1763265175272329
                    environment_marker=package_attributes.get("marker"),
                )

                packages.append(package)
            else:
                # Despite we can continue the loop ignoring this particular wheel,
                # we should raise an exception to make us be aware of this issue in case
                # it's hiding other issues.
                raise CondaException(
                    "We encountered a wheel package that's missing an url or a hash. "
                    + f"Please escalate this case to the Metaflow team. url={url}, file_hashes={file_hashes}"
                )

        return packages

    # TODO(https://netflix.atlassian.net/browse/MLPMF-502): use the open source PyLock
    # code to parse the toml file instead of parsing it to a dict as in current code.
    @staticmethod
    def _pylock_toml_root_obj_to_packages(
        toml_root_obj: Dict[str, Any],
    ) -> Dict[str, List[PackageSpecification]]:
        # Despite pylock.toml being a standard format, let's refrain the scope of
        # of supported scenarios to reduce risks.
        if toml_root_obj.get("created-by", "") != "uv":
            raise CondaException(
                "Currently we only support pylock.toml file generated by uv. Please contact "
                "the Metaflow team if you need us to support more general cases."
            )

        packages = toml_root_obj.get("packages", [])
        out_dict: Dict[str, List[PackageSpecification]] = {}

        for package in packages:
            if not isinstance(package, dict):
                raise CondaException(
                    "Each package in 'packages' should be a dictionary."
                )
            name = package.get("name", "")
            if not name:
                # https://peps.python.org/pep-0751/#packages-name: name is required.
                raise CondaException("Each package in 'packages' must have a name.")

            # Packages of the same name can appear multiple times with different markers.
            # We will apply filtering of them later based on the environment markers.
            # See https://netflix.atlassian.net/browse/MLPMF-545.
            out_dict.setdefault(name, []).extend(
                PylockTomlResolver._toml_package_to_package_specs(package)
            )

        return out_dict

    @staticmethod
    def _packages_to_resolved_env(
        packages: List[PackageSpecification],
        env_type: EnvType,
        deps: Dict[str, List[str]],
        sources: Optional[Dict[str, List[str]]],
        extras: Optional[Dict[str, List[str]]],
        architecture: Optional[str] = None,
    ):
        resolved_env = ResolvedEnvironment(
            deps,
            sources,
            extras,
            architecture,
            all_packages=packages,
            env_type=env_type,
        )

        return resolved_env

    """
    Get the best compatible packages for the given supported tags.

    STEP 1 (Context):
    
    The full list of available packages are generated by:
    1. Filtering out packages that don't satisfy the user's conditions (such as by version specifiers)
       For example, "numpy<2.0"
    2. Group all packages by their names.

    For example, for package name = "charset-normalizer", we could the following potential packages:
    { url = "https://pypi.netflix.net/packages/18953191400/charset_normalizer-3.4.3-cp310-cp310-macosx_10_9_universal2.whl", upload-time = 2025-08-09T07:55:36Z, size = 207695, hashes = { sha256 = "fb7f67a1bfa6e40b438170ebdc8158b78dc465a5a67b6dde178a46987b244a72" } },
    { url = "https://pypi.netflix.net/packages/18953191401/charset_normalizer-3.4.3-cp310-cp310-manylinux2014_aarch64.manylinux_2_17_aarch64.manylinux_2_28_aarch64.whl", upload-time = 2025-08-09T07:55:38Z, size = 147153, hashes = { sha256 = "cc9370a2da1ac13f0153780040f465839e6cccb4a1e44810124b4e22483c93fe" } },
    { url = "https://pypi.netflix.net/packages/18953191402/charset_normalizer-3.4.3-cp310-cp310-manylinux2014_ppc64le.manylinux_2_17_ppc64le.manylinux_2_28_ppc64le.whl", upload-time = 2025-08-09T07:55:40Z, size = 160428, hashes = { sha256 = "07a0eae9e2787b586e129fdcbe1af6997f8d0e5abaa0bc98c0e20e124d67e601" } },
    { url = "https://pypi.netflix.net/packages/18953191403/charset_normalizer-3.4.3-cp310-cp310-manylinux2014_s390x.manylinux_2_17_s390x.manylinux_2_28_s390x.whl", upload-time = 2025-08-09T07:55:41Z, size = 157627, hashes = { sha256 = "74d77e25adda8581ffc1c720f1c81ca082921329452eba58b16233ab1842141c" } },
    { url = "https://pypi.netflix.net/packages/18953191404/charset_normalizer-3.4.3-cp310-cp310-manylinux2014_x86_64.manylinux_2_17_x86_64.manylinux_2_28_x86_64.whl", upload-time = 2025-08-09T07:55:43Z, size = 152388, hashes = { sha256 = "d0e909868420b7049dafd3a31d45125b31143eec59235311fc4c57ea26a4acd2" } },
    { url = "https://pypi.netflix.net/packages/18953191405/charset_normalizer-3.4.3-cp310-cp310-musllinux_1_2_aarch64.whl", upload-time = 2025-08-09T07:55:44Z, size = 150077, hashes = { sha256 = "c6f162aabe9a91a309510d74eeb6507fab5fff92337a15acbe77753d88d9dcf0" } },
    { url = "https://pypi.netflix.net/packages/18953191406/charset_normalizer-3.4.3-cp310-cp310-musllinux_1_2_ppc64le.whl", upload-time = 2025-08-09T07:55:46Z, size = 161631, hashes = { sha256 = "4ca4c094de7771a98d7fbd67d9e5dbf1eb73efa4f744a730437d8a3a5cf994f0" } },
    { url = "https://pypi.netflix.net/packages/18953191407/charset_normalizer-3.4.3-cp310-cp310-musllinux_1_2_s390x.whl", upload-time = 2025-08-09T07:55:47Z, size = 159210, hashes = { sha256 = "02425242e96bcf29a49711b0ca9f37e451da7c70562bc10e8ed992a5a7a25cc0" } },

    STEP 2 (Context):

    Then we calculated the supported_tags by calling pypi_tags_from_arch(), which generates
    a list of tags that are compatible with the current architecture.
    An example of the supported tags:
    https://github.netflix.net/gist/shuishiy/af1f3287731931b443fddecb005fb7a6

    STEP 3 (this method):
    
    This method iterates through all the supported tags by the current architecture,
    and iterates all the packages in its original sorted order. For each package, the
    architecture is parsed through its package url or file name. And the first package
    that satisfies the current architecture is returned. There is an underlying assumption
    that the package is already ordered in a desired way.
    """



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false

from typing import Dict, List, Optional, Tuple

from metaflow.debug import debug

from ..env_descr import EnvType, ResolvedEnvironment
from ..utils import arch_id, get_builder_envs_dep
from . import Resolver
from .conda_resolver import CondaResolver


class BuilderEnvsResolver(Resolver):
    TYPES = ["builderenv"]

    def resolve(
        self,
        env_type: EnvType,
        python_version_requested: str,
        deps: Dict[str, List[str]],
        sources: Dict[str, List[str]],
        extras: Dict[str, List[str]],
        architecture: str,
        builder_envs: Optional[List[ResolvedEnvironment]] = None,
        base_env: Optional[ResolvedEnvironment] = None,
        file_paths: Dict[str, List[str]] = {},
    ) -> Tuple[ResolvedEnvironment, Optional[List[ResolvedEnvironment]]]:
        conda_only_sources = sources.get("conda", [])
        python_deps = [d for d in deps.get("conda", []) if d.startswith("python==")]
        # We add a few more packages that we need to support building wheels
        # Conda typically includes pip but no harm adding it there too
        # All packages included are miniscule and have no dependencies.
        python_deps.extend(get_builder_envs_dep(deps.get("conda", [])))

        if arch_id() == architecture:
            conda_only_deps = {
                "npconda": deps.get("npconda", []),
                "sys": deps.get("sys", []),
                "conda": python_deps,
            }
            debug.conda_exec(
                "Building builder environment with %s" % str(conda_only_deps)
            )
            return CondaResolver(self._conda).resolve(
                EnvType.CONDA_ONLY,
                python_version_requested,
                conda_only_deps,
                {"conda": conda_only_sources},
                extras,
                architecture,
                builder_envs,
                None,
            )

        debug.conda_exec("Using vanilla builder env with %s" % str(python_deps[0]))

        return CondaResolver(self._conda).resolve(
            EnvType.CONDA_ONLY,
            python_version_requested,
            {"conda": python_deps},
            {"conda": conda_only_sources},
            extras,
            arch_id(),
            builder_envs,
            None,
        )



_jupyter_title_markdown = """
# Debugging Notebook for Task: {TaskPathSpec}

### Debug Type: {DebugType}

- A `Pre-Execution State` debug type implies that you can inspect the state of 
the task before it began execution. 
- For instance, if the current Metaflow step is called `stepB` and the 
previous step was `stepA`, then you are essentially debugging the state 
after `stepA` finished execution and before `stepB` began execution.
- In case of a `Post-Execution State` debug type, you can inspect the state 
of the task after it has finished execution. In this case, you can inspect 
the state of the task after `stepB` has finished execution, and before any 
succeeding step has started.
"""



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false
import json
import os
import shutil
import sys
import time
from typing import Any

from metaflow.metaflow_config import DATASTORE_LOCAL_DIR, CONDA_MAGIC_FILE_V2

from metaflow.cli import echo_always

from metaflow.debug import debug

from metaflow.packaging_sys import ContentType, MetaflowCodeContent
from metaflow.plugins.env_escape import generate_trampolines, ENV_ESCAPE_PY

from .conda import Conda
from .conda_environment import CondaEnvironment
from .env_descr import EnvID
from .utils import arch_id, plural_marker


def my_echo_always(*args: Any, **kwargs: Any) -> None:
    kwargs["err"] = False
    echo_always(*args, **kwargs)


def bootstrap_environment(
    flow_name: str, step_name: str, req_id: str, full_id: str, datastore_type: str
):
    start = time.time()
    my_echo_always("    Setting up Conda ...", nl=False)
    setup_conda_manifest()
    my_conda = Conda(my_echo_always, datastore_type, mode="remote")
    # Access a binary to force the downloading of the remote conda
    my_conda.binary("micromamba")
    delta_time = int(time.time() - start)
    my_echo_always(" done in %d second%s." % (delta_time, plural_marker(delta_time)))

    # Resolve a late environment if full_id is a special string _fetch_exec
    if full_id == "_fetch_exec":
        alias_to_fetch = CondaEnvironment.sub_envvars_in_envname(req_id)
        env_id = my_conda.env_id_from_alias(alias_to_fetch)
        if env_id is None:
            raise RuntimeError(
                "Cannot find environment '%s' (from '%s') for arch '%s'"
                % (alias_to_fetch, req_id, arch_id())
            )
        req_id = env_id.req_id
        full_id = env_id.full_id
    resolved_env = my_conda.environment(
        EnvID(req_id=req_id, full_id=full_id, arch=arch_id())
    )
    if resolved_env is None:
        raise RuntimeError(
            "Cannot find cached environment for hash %s:%s" % (req_id, full_id)
        )
    # Install the environment; this will fetch packages as well.
    python_bin = os.path.join(
        my_conda.create_for_step(step_name, resolved_env, do_symlink=True),
        "bin",
        "python",
    )
    # Setup anything needed by the escape hatch
    if ENV_ESCAPE_PY is not None:
        cwd = os.getcwd()
        trampoline_dir = os.path.join(cwd, "_escape_trampolines")
        os.makedirs(trampoline_dir)
        generate_trampolines(trampoline_dir)
        # print("Environment escape will use %s as the interpreter" % ENV_ESCAPE_PY)
    else:
        pass
        # print("Could not find a environment escape interpreter")
    # We write out the env_id to _env_id so it can be read by the outer bash script
    with open("_env_id", mode="w", encoding="utf-8") as f:
        json.dump(EnvID(req_id, full_id, arch_id()), f)

    # Same thing for env path (used to set lib for example)
    with open("_env_path", mode="w", encoding="utf-8") as f:
        if python_bin is None:
            raise RuntimeError("Environment was not created properly")
        json.dump(os.path.dirname(os.path.dirname(python_bin)), f)


def setup_conda_manifest():
    manifest_folder = os.path.join(os.getcwd(), DATASTORE_LOCAL_DIR)
    if not os.path.exists(manifest_folder):
        os.makedirs(manifest_folder)
    path_to_manifest = MetaflowCodeContent.get_filename(
        CONDA_MAGIC_FILE_V2, ContentType.OTHER_CONTENT
    )
    if path_to_manifest is None:
        raise RuntimeError(
            "Cannot find the conda manifest file %s in the package"
            % CONDA_MAGIC_FILE_V2
        )
    shutil.move(
        path_to_manifest,
        os.path.join(manifest_folder, CONDA_MAGIC_FILE_V2),
    )


if __name__ == "__main__":
    start = time.time()
    bootstrap_environment(*sys.argv[1:])
    debug.conda_exec("Conda bootstrap took %f seconds" % (time.time() - start))



"""HTTP helpers for Netflix extensions."""

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


def create_http_adapter_with_retry(
    retries=3,
    backoff_factor=0.3,
    status_forcelist=(500, 502, 503, 504),
    **kwargs
):
    """
    Create an HTTP adapter with retry logic.

    Parameters
    ----------
    retries : int
        Number of retries
    backoff_factor : float
        Backoff factor for retries
    status_forcelist : tuple
        HTTP status codes to retry on
    **kwargs : dict
        Additional keyword arguments for the adapter

    Returns
    -------
    HTTPAdapter
        HTTP adapter with retry logic
    """
    retry_strategy = Retry(
        total=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
        **kwargs
    )
    adapter = HTTPAdapter(max_retries=retry_strategy)
    return adapter



import json
import os
import platform
import re
import tarfile

from io import BytesIO
from itertools import chain
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,  # noqa
    Tuple,
    Union,
    cast,
)

from metaflow.debug import debug

from metaflow.packaging_sys import ContentType
from metaflow.plugins.datastores.local_storage import LocalStorage
from metaflow.flowspec import FlowSpec

from metaflow.metaflow_config import (
    CONDA_MAGIC_FILE_V2,  # type: ignore[attr-defined]
    CONDA_REMOTE_COMMANDS,  # type: ignore[attr-defined]
    CONDA_TEST,  # type: ignore[attr-defined]
    CONDA_IGNORE_CACHING_DATASTORES,  # type: ignore[attr-defined]
    get_pinned_conda_libs,
)

from metaflow.metaflow_environment import (
    InvalidEnvironmentException,
    MetaflowEnvironment,
)

from metaflow.unbounded_foreach import UBF_TASK

from metaflow._vendor.packaging.utils import canonicalize_version

from .envsresolver import EnvsResolver
from .utils import (
    arch_id,
    channel_or_url,
    conda_deps_to_pypi_deps,
    get_conda_manifest_path,
    get_sys_packages,
    merge_dep_dicts,
    resolve_env_alias,
    sanitize_python_version,
)

from .env_descr import (
    AliasType,
    CachedEnvironmentInfo,
    EnvID,
    EnvType,
    ResolvedEnvironment,
)
from .conda import Conda

from .conda_common_decorator import StepRequirement, StepRequirementMixin


class CondaEnvironment(MetaflowEnvironment):
    TYPE = "conda"
    _filecache = None
    _conda = None
    _flow_req = None  # type: Optional[StepRequirement]
    _result_for_step = (
        {}
    )  # type: Dict[str, Tuple[str, StepRequirement, Optional[Tuple[EnvID, EnvType, Optional[ResolvedEnvironment]]]]]

    def __init__(self, flow: FlowSpec):
        self._flow = flow

        self.conda = None  # type: Optional[Conda]

        # A conda environment sits on top of whatever default environment
        # the user has so we get that environment to be able to forward
        # any calls we don't handle specifically to that one.
        from metaflow.plugins import ENVIRONMENTS
        from metaflow.metaflow_config import DEFAULT_ENVIRONMENT

        if DEFAULT_ENVIRONMENT == self.TYPE:
            # If the default environment is Conda itself then fallback on
            # the default 'default environment'
            self.base_env = MetaflowEnvironment(self._flow)
        else:
            self.base_env = [
                e
                for e in ENVIRONMENTS + [MetaflowEnvironment]
                if e.TYPE == DEFAULT_ENVIRONMENT
            ][0](self._flow)

    def init_environment(self, echo: Callable[..., None]):
        # Print a message for now
        echo("Bootstrapping Conda environment... (this could take a few minutes)")

        self.conda = cast(Conda, self.conda)
        resolver = EnvsResolver(self.conda)

        for step in self._flow:
            # Figure out the environments that we need to resolve for all steps
            # We will resolve all unique environments in parallel
            _, arch, req, base_env = self.extract_merged_reqs_for_step(
                self.conda, self._flow, self._datastore_type, step
            )

            if req.is_disabled or req.is_fetch_at_exec:
                continue

            if req.from_env_name and base_env is None:
                raise InvalidEnvironmentException(
                    "Base environment '%s' was not found for architecture '%s'"
                    % (req.from_env_name, arch)
                )
            # Determine if the base is a full-id -- we don't currently allow this so
            # this should always be false but keeping to not forget later
            base_from_full_id = False
            if req.from_env_name:
                base_from_full_id = (
                    resolve_env_alias(req.from_env_name)[0] == AliasType.FULL_ID
                )

            resolver.add_environment(
                arch,
                req.packages_as_str,
                req.sources,
                {},
                req.file_paths,
                step.name,
                base_env,
                base_from_full_id=base_from_full_id,
                full_id_unique_keys=req.full_id_unique_keys,
            )

        resolver.resolve_environments(echo)
        self._log_conda_events(resolver)

        update_envs = []  # type: List[ResolvedEnvironment]
        if self._datastore_type not in CONDA_IGNORE_CACHING_DATASTORES or CONDA_TEST:
            # We may need to update caches
            # Note that it is possible that something we needed to resolve, we don't need
            # to cache (if we resolved to something already cached).
            formats = set()  # type: Set[str]
            for _, resolved_env, f, _ in resolver.need_caching_environments():
                update_envs.append(resolved_env)
                formats.update(f)

            self.conda.cache_environments(update_envs, {"conda": list(formats)})
        else:
            update_envs = [
                resolved_env for _, resolved_env, _ in resolver.new_environments()
            ]

        self.conda.add_environments(update_envs)

        # Update the default environment
        for env_id, resolved_env, _ in resolver.resolved_environments():
            if resolved_env and env_id.full_id == "_default":
                self.conda.set_default_environment(resolved_env.env_id)

        # We are done -- write back out the environments.
        # TODO: Not great that this is manual
        self.conda.write_out_environments()

        # Delegate to whatever the base environment needs to do.
        self.base_env.init_environment(echo)

    def validate_environment(
        self, echo: Callable[..., Any], datastore_type: str
    ) -> None:
        self._local_root = cast(str, LocalStorage.get_datastore_root_from_config(echo))
        self._datastore_type = datastore_type
        self.conda = Conda(echo, datastore_type)

        return self.base_env.validate_environment(echo, datastore_type)

    def decospecs(self) -> Tuple[str, ...]:
        return ("conda_env_internal",) + self.base_env.decospecs()

    def bootstrap_commands(
        self,
        step_name: str,
        datastore_type: str,
    ) -> List[str]:
        # Bootstrap conda and execution environment for step
        env_id = self.get_env_id_noconda(step_name)
        if env_id is not None:
            if isinstance(env_id, EnvID):
                arg1 = env_id.req_id
                arg2 = env_id.full_id
            else:
                arg1 = env_id
                arg2 = "_fetch_exec"
            return [
                "export CONDA_START=$(date +%s)",
                "echo 'Bootstrapping environment ...'",
                'python -m %s.remote_bootstrap "%s" "%s" %s %s %s'
                % (
                    "metaflow_extensions.nflx.plugins.conda",
                    self._flow.name,
                    step_name,
                    arg1,
                    arg2,
                    datastore_type,
                ),
                "export _METAFLOW_CONDA_ENV=$(cat _env_id)",
                "export PYTHONPATH=$(pwd)/_escape_trampolines:$(printenv PYTHONPATH)",
                # NOTE: Assumes here that remote nodes are Linux
                "if [[ -n $(printenv LD_LIBRARY_PATH) ]]; then "
                "export MF_ORIG_LD_LIBRARY_PATH=$(printenv LD_LIBRARY_PATH); "
                "export LD_LIBRARY_PATH=$(cat _env_path)/lib:$(printenv LD_LIBRARY_PATH); fi",
                "echo 'Environment bootstrapped.'",
                "export CONDA_END=$(date +%s)",
            ]
        return []

    def add_to_package(self) -> List[Tuple[str, str]]:
        # TODO: Improve this to only extract the environments that we care about
        # The issue is that we need to pass a path which is a bit annoying since
        # we then can't clean it up easily.
        files = self.base_env.add_to_package()
        # Add conda manifest file to job package at the top level.

        # In the case of a resume, we still "package" the code but don't send it
        # anywhere -- if we are resuming on a remote node, this file may not exist
        # so we skip
        path = get_conda_manifest_path(self._local_root)
        if path and os.path.exists(path):
            files.append((path, os.path.basename(path), ContentType.OTHER_CONTENT))
        return files

    def pylint_config(self) -> List[str]:
        config = self.base_env.pylint_config()
        # Disable (import-error) in pylint
        config.append("--disable=F0401")
        return config

    @classmethod
    def get_client_info(
        cls, flow_name: str, metadata: Dict[str, str]
    ) -> Dict[str, Any]:
        # TODO: FIX THIS: this doesn't work with the new format.
        if cls._filecache is None:
            from metaflow.client.filecache import FileCache

            cls._filecache = FileCache()
        info_str = metadata.get("code-package")
        env_id = json.loads(metadata.get("conda_env_id", "[]"))
        if info_str is None or not env_id:
            return {"type": "conda"}
        info = json.loads(info_str)
        if info is None:
            return {"type": "conda"}
        _, blobdata = cls._filecache.get_data(
            info["ds_type"], flow_name, info["location"], info["sha"]
        )
        with tarfile.open(fileobj=BytesIO(blobdata), mode="r:gz") as tar:
            conda_file = tar.extractfile(CONDA_MAGIC_FILE_V2)
        if conda_file is None:
            return {"type": "conda"}
        cached_env_info = CachedEnvironmentInfo.from_dict(
            json.loads(conda_file.read().decode("utf-8"))
        )
        resolved_env = cached_env_info.env_for(env_id[0], env_id[1])
        if resolved_env:
            new_info = {
                "type": "conda",
                "req_id": env_id[0],
                "full_id": env_id[1],
                "user_deps": "; ".join(map(str, resolved_env.deps)),
                "all_packages": "; ".join(
                    [
                        p.filename
                        for p in sorted(resolved_env.packages, key=lambda p: p.filename)
                    ]
                ),
            }
            return new_info
        return {"type": "conda"}

    def get_package_commands(
        self,
        code_package_url: str,
        datastore_type: str,
        code_package_metadata: Optional[str] = None,
    ):
        return self.base_env.get_package_commands(
            code_package_url, datastore_type, code_package_metadata
        )

    def get_environment_info(self, include_ext_info=False):
        return self.base_env.get_environment_info(include_ext_info)

    def executable(self, step_name: str, default: Optional[str] = None) -> str:
        # Get relevant python interpreter for step
        executable = self._get_executable(step_name)
        if executable is not None:
            return executable
        return self.base_env.executable(step_name, default)

    # The below methods are specific to this class (not from the base class)
    def get_env_id_noconda(self, step_name: str) -> Optional[Union[str, EnvID]]:
        return self.get_env_id(cast(Conda, self.conda), step_name)

    def resolve_fetch_at_exec_env(
        self,
        step_name: str,
        envvars: Optional[Dict[str, Union[str, Callable[[], str]]]] = None,
    ) -> Optional[EnvID]:
        step_result = self._result_for_step.get(step_name)
        if step_result is None or step_result[1].from_env_name is None:
            return None
        resolved_name = self.sub_envvars_in_envname(
            step_result[1].from_env_name, envvars
        )

        resolved_env_id = cast(Conda, self.conda).env_id_from_alias(
            resolved_name, arch=step_result[0]
        )
        if resolved_env_id is None:
            raise RuntimeError(
                "Cannot find environment '%s' (from '%s') for arch '%s'"
                % (resolved_name, step_result[1].from_env_name, step_result[0])
            )
        return resolved_env_id

    @classmethod
    def get_env_id(cls, conda: Conda, step_name: str) -> Optional[Union[str, EnvID]]:
        # Returns the environment ID for the step:
        #  - if the step has conda disabled or has no known resolved env, it returns None
        #  - if the step has fetch_at_exec, it returns the environment name to fetch on
        #    execution
        #  - if the step has a conda environment, it returns the environment ID
        #
        # We either look in _result_for_step for the env id (this only works for the
        # initial process that launched Metaflow ie: the first CLI process). This is what
        # happens when deploying to argo/airflow or any other scheduler. For everything
        # else, we look at _METAFLOW_CONDA_ENV:
        #  - when deploying to a scheduler: use _result_for_step to build the command line
        #  - when executing for scheduler, _METAFLOW_CONDA_ENV is set by remote_bootstrap
        #    and will also fully resolve fetch_at_exec and used in task_pre_step. Note that
        #    it relies on the value that was hardcoded in the command line when deploying
        #    to the scheduler. The value may still not be the same due to fetch_at_exec
        #  - when using the runtime:
        #    - if executing locally, the environment is created in runtime_step_cli
        #      and sets _METAFLOW_CONDA_ENV which is then read in task_pre_step (in the
        #      next process)
        #    - if executing remotely, _METAFLOW_CONDA_ENV is set by runtime_step_cli. It
        #      is then used by bootstrap_commands (in conda_environment.py) to build the
        #      command line. The command line will then call remote_bootstrap which will
        #      set again _METAFLOW_CONDA_ENV to the fully resolved environment which is
        #      then used in task_pre_step.
        step_info = cls._result_for_step.get(step_name)
        if step_info is None:
            resolved_env_id = None
            t = os.environ.get("_METAFLOW_CONDA_ENV")
            if t:
                resolved_env_id = EnvID(*json.loads(t))
            return resolved_env_id
        elif step_info[1].is_disabled:
            # Another case for the use of this function is the "show" command so
            # in that case we do not have _METAFLOW_CONDA_ENV but we do have the
            # information in step_info[1]
            return None
        elif step_info[1].is_fetch_at_exec:
            return step_info[1].from_env_name
        elif step_info[2]:
            # In this case, we should know about the environment -- it will have
            # _default flag at this time
            resolved_env = conda.environment(
                step_info[2][0],
                local_only=True,
                full_id_unique_keys=step_info[1].full_id_unique_keys,
            )
            if resolved_env:
                return resolved_env.env_id
            return None
        else:
            # Default case when step_info exists but no conditions are met
            return None

    @classmethod
    def enabled_for_step(cls, step_name: str, ubf_context: str) -> bool:
        step_info = cls._result_for_step.get(step_name)
        if step_info:
            if step_info[1].is_disabled is None:
                if step_info[1].default_disabled(ubf_context) is None:
                    return ubf_context == UBF_TASK
                return not step_info[1].default_disabled(ubf_context)
            return not step_info[1].is_disabled
        return os.environ.get("_METAFLOW_CONDA_ENV") is not None

    @classmethod
    def fetch_at_exec_for_step(cls, step_name: str) -> bool:
        # NOTE: This method does not work after runtime_step_cli
        # The good thing is we don't need it :)
        step_info = cls._result_for_step.get(step_name)
        if step_info:
            return step_info[1].is_fetch_at_exec or False
        return False

    @classmethod
    def unresolved_from_name_for_step(cls, step_name: str) -> Optional[str]:
        step_info = cls._result_for_step.get(step_name)
        if step_info:
            return step_info[1].from_env_name
        return None

    @classmethod
    def extract_reqs_for_flow(cls, flow: FlowSpec) -> StepRequirement:
        if cls._flow_req is not None:
            return cls._flow_req.copy()

        cls._flow_req = StepRequirement()
        for decos in flow._flow_decorators.values():
            for deco in decos:
                if isinstance(deco, StepRequirementMixin):
                    cls._flow_req.merge_update(deco)
        return cls._flow_req.copy()

    @classmethod
    def extract_reqs_for_step(cls, step: Any) -> StepRequirement:
        req = StepRequirement()
        for step_deco in step.decorators:
            if isinstance(step_deco, StepRequirementMixin):
                req.merge_update(step_deco)
        return req

    @classmethod
    def extract_merged_reqs_for_step(
        cls,
        conda: Conda,
        flow: FlowSpec,
        datastore_type: str,
        step: Any,
        override_arch: Optional[str] = None,
        local_only: bool = False,
    ) -> Tuple[EnvType, str, StepRequirement, Optional[ResolvedEnvironment]]:
        # We typically only have one architecture per step but it is possible to
        # use the resolve CLI and pass multiple architectures. In that case, we use
        # override_arch and so in that case, we skip caching.
        computed_result = cls._result_for_step.get(step.name)
        if override_arch is None and computed_result:
            if computed_result[2] is None:
                # Disabled or fetch_at_exec
                return EnvType.CONDA_ONLY, computed_result[0], computed_result[1], None
            return (
                computed_result[2][1],
                computed_result[0],
                computed_result[1],
                computed_result[2][2],
            )

        # First figure out the architecture for this step as well as if a GPU
        # is needed (to inject the proper __cuda dependency)
        resources_deco = [
            deco
            for deco in step.decorators
            if deco.name in ("resources", *CONDA_REMOTE_COMMANDS)
        ]
        step_is_remote = False
        step_gpu_requested = False
        for deco in resources_deco:
            if deco.name in CONDA_REMOTE_COMMANDS:
                step_is_remote = True
            if deco.attributes.get("gpu") not in (None, 0, "0"):
                step_gpu_requested = True

        if override_arch:
            step_arch = override_arch
        else:
            step_arch = "linux-64" if step_is_remote else arch_id()

        final_req = cls.extract_reqs_for_flow(flow)
        step_req = cls.extract_reqs_for_step(step)

        debug.conda_exec(
            "For step %s, got flow req: %s; step req: %s"
            % (step.name, repr(final_req), repr(step_req))
        )

        final_req.override_update(step_req)

        debug.conda_exec(
            "For step %s, merged requirement: %s" % (step.name, repr(final_req))
        )

        if not computed_result and final_req.is_disabled:
            if override_arch is None:
                cls._result_for_step[step.name] = (step_arch, final_req, None)
            return (
                EnvType.CONDA_ONLY,
                step_arch,
                final_req,
                None,
            )  # No point going further -- it is disabled
        if final_req.is_fetch_at_exec:
            # In this case, we check things are valid
            if final_req.from_env_name is None:
                raise InvalidEnvironmentException(
                    "In step '%s', a 'fetch-at-exec' environment needs to have an "
                    "environment specified with '@named_env'" % step.name
                )
            if (
                any([v for v in final_req.sources.values()])
                or any([v for v in final_req.packages.values()])
                or final_req.python
            ):
                raise InvalidEnvironmentException(
                    "In step '%s', a 'fetch_at_exec' environment cannot have "
                    "any additional requirements (packages, sources, or python)"
                    % step.name
                )
            if override_arch is None:
                cls._result_for_step[step.name] = (step_arch, final_req, None)
            return (
                EnvType.CONDA_ONLY,
                step_arch,
                final_req,
                None,  # No point going further -- we won't be able to get all info
            )

        all_packages: Dict[str, Dict[str, str]] = final_req.packages

        # Figure out the from_env information
        from_env_name = final_req.from_env_name

        from_env = None
        if from_env_name:
            from_env_id = conda.env_id_from_alias(
                from_env_name, step_arch, local_only=local_only
            )
            if from_env_id is not None:
                from_env = conda.environment(from_env_id, local_only=local_only)
                if from_env is None:
                    raise InvalidEnvironmentException(
                        "Base environment '%s' exists but is not available for architecture '%s'"
                        % (from_env_name, step_arch)
                    )
            else:
                raise InvalidEnvironmentException(
                    "Base environment '%s' not found" % from_env_name
                )

        env_type = CondaEnvironment._determine_env_type_for_step(from_env, final_req)

        if from_env is not None:
            for p in from_env.packages:
                if p.package_name == "python":
                    final_req.python = p.package_version
                    break
            else:
                raise InvalidEnvironmentException(
                    "Cannot determine Python version from the base environment"
                )
        else:
            if final_req.python is None:
                final_req.python = sanitize_python_version(platform.python_version())
            # Ensure final_req.python is not None before calling canonicalize_version
            python_version = final_req.python
            if python_version is not None:
                all_packages.setdefault("conda", {})["python"] = canonicalize_version(
                    python_version
                )

        # Add pinned dependencies based on env-type; we prefer conda dependencies if
        # env-type is mixed

        if env_type == EnvType.PYPI_ONLY:
            all_packages["pypi"] = merge_dep_dicts(
                all_packages.get("pypi", {}),
                conda_deps_to_pypi_deps(
                    get_pinned_conda_libs(final_req.python, datastore_type)
                ),
            )
        else:
            all_packages["conda"] = merge_dep_dicts(
                all_packages.get("conda", {}),
                get_pinned_conda_libs(final_req.python, datastore_type),
            )

        # Add the system requirements and default channels.
        # The default channels go into the computation of the req ID so it is important
        # to have them at this time.

        sys_pkgs = get_sys_packages(
            conda.virtual_packages, step_arch, step_gpu_requested
        )

        # The user can specify whatever they want but we inject things they don't
        # specify
        final_req_sys = all_packages.setdefault("sys", {})
        for pkg_name, pkg_version in sys_pkgs.items():
            if pkg_name not in final_req_sys:
                final_req_sys[pkg_name] = pkg_version

        final_req._packages = all_packages

        # Update sources -- here the order is important so we explicitly set it
        # This code will put:
        #  - unique conda sources using user sources *first*
        #  - conda sources will be just channels if possible
        #  - pypi sources: default ones first followed by users who basically adds
        #    only extra-indices
        final_sources = {
            "conda": list(
                dict.fromkeys(
                    map(
                        channel_or_url,
                        chain(
                            final_req.sources.get("conda", []),
                            conda.default_conda_channels,
                        ),
                    )
                )
            )
        }

        if env_type != EnvType.CONDA_ONLY:
            final_sources["pypi"] = list(
                dict.fromkeys(
                    chain(conda.default_pypi_sources, final_req.sources.get("pypi", []))
                )
            )

        final_req.sources = final_sources

        debug.conda_exec("For step %s, final req: %s" % (step.name, repr(final_req)))
        # TODO: This is a bit wasteful because we will do this twice but we need
        # to store the env_id here in case this is called from the CLI (ie: we need
        # to be able to get the req_id from steps even if init_environment is not called)
        # We could improve this though it is likely not a huge overhead.
        if from_env:
            _, env_id, _, _, _, _, _ = EnvsResolver.extract_info_from_base(
                conda,
                from_env,
                final_req.packages_as_str,
                final_req.sources,
                {},
                step_arch,
            )
        else:
            env_id = EnvID(
                ResolvedEnvironment.get_req_id(
                    final_req.packages_as_str, final_req.sources, {}
                ),
                "_default",
                step_arch,
            )
        if override_arch is None:
            cls._result_for_step[step.name] = (
                step_arch,
                final_req,
                (env_id, env_type, from_env),
            )

        return env_type, step_arch, final_req, from_env

    # Determines the environment type for a step given its final requirements.
    #
    # There are two "sources" of information that determine the environment type:
    # 1) The env_type of the from_env (if any)
    # 2) The presence of conda/pypi packages in the final requirements
    # If
    # a) from_env is CONDA_ONLY and only has conda packages --> return CONDA_ONLY
    # b) from_env is PYPI_ONLY and only has pypi packages --> return PYPI_ONLY
    # c) presence of two different things, such as a CONDA_ONLY from_env and a pypi package
    #    presence --> return MIXED.
    @staticmethod
    def _determine_env_type_for_step(
        from_env: Optional[ResolvedEnvironment],
        # Deliberately taking a StepRequirement rather than all_packages,
        # in case later we want to consider other fields.
        final_req: StepRequirement,
    ) -> EnvType:

        from_env_env_type: Optional[EnvType] = (
            None if from_env is None else from_env.env_type
        )

        # Use the private variables directly to avoid deep copy on packages(), sources() properties.
        has_conda = len(final_req._packages.get("conda", {})) > 0
        has_pypi = len(final_req._packages.get("pypi", {})) > 0

        mapping = dict[EnvType, bool](
            {
                EnvType.CONDA_ONLY: has_conda,
                EnvType.PYPI_ONLY: has_pypi,
            }
        )

        env_hits = sum(
            # from_env env_type is the same as "has__", then that counts as one hit (bool true->1)
            int(k == from_env_env_type or v)
            for (k, v) in mapping.items()
        )

        if env_hits > 1:
            return EnvType.MIXED
        elif env_hits == 1:
            # env_hits == 1 is only possible in the following cases:
            # (a) from_env_env_type is the same as the only present package, such as
            #     from_env_env_type == CONDA_ONLY and has_conda == True and has_pypi == False
            # (b) from_env_env_type is None and only one of has_conda or has_pypi is True,
            #     in this case we return the one that is True
            # (c) from_env_env_type is MIXED and only one package presence is counted,
            #     such as has_conda == True and has_pypi == False or vice-versa,
            #     in this case we return MIXED.
            return (
                from_env_env_type
                # When from_env_env_type is not None, we have (a) or (b) scenarios in above.
                # In those two cases, we return the same from_env_env_type.
                if from_env_env_type is not None
                else (
                    # When from_env_env_type is None, we have (b) scenario in above.
                    next(k for k, v in mapping.items() if v)
                )
            )
        else:
            # It should be be impossible to have both has_conda==False and has_pypi==False,
            # and this line of code shouldn't be reached.
            # But just to be safe for this refactoring PR, we keep the original behavior.
            # In the prev version of this file, when two lines at
            # https://github.netflix.net/corp/mli-metaflow-custom/blob/051bb886e99f61ba1de1eb909cd31379f4d87c75/nflx-metaflow/metaflow_extensions/nflx/plugins/conda/conda_environment.py#L524
            # get both False, and line
            # L547
            # assigns env_type = EnvType.CONDA_ONLY,
            # and L566, L568 are not hit, the eventual env_type is EnvType.CONDA_ONLY.
            return EnvType.CONDA_ONLY  # Default to conda only

    @staticmethod
    def sub_envvars_in_envname(
        name: str, addl_env: Optional[Dict[str, Union[str, Callable[[], str]]]] = None
    ) -> str:
        if addl_env is None:
            addl_env = {}
        envvars_to_sub = re.findall(r"\@{(\w+)}", name)
        for envvar in set(envvars_to_sub):
            replacement = os.environ.get(envvar, addl_env.get(envvar))
            if callable(replacement):
                replacement = replacement()
            if replacement is not None:
                name = name.replace("@{%s}" % envvar, replacement)
            else:
                raise InvalidEnvironmentException(
                    "Could not find '%s' in the environment -- needed to resolve '%s'"
                    % (envvar, name)
                )
        return name

    def _get_executable(self, step_name: str) -> Optional[str]:
        env_id = self.get_env_id_noconda(step_name)
        if env_id is not None:
            # The create method in Conda() sets up this symlink when creating the
            # environment.
            return os.path.join(".", "__conda_python")
        return None

    def _log_conda_events(self, resolver):
        from metaflow.system import _system_logger

        def _format_packages(packages):
            return {
                package.package_name: {
                    "package_version": package.package_version,
                    "filename": package.filename,
                    "package_url": package.url,
                    "package_detailed_version": package.package_detailed_version,
                    "package_format": package.url_format,
                }
                for package in packages
            }

        for (
            newly_resolved_env_id,
            newly_resolved_env,
            needed_in_steps,
        ) in resolver.new_environments():
            # Log the requested packages
            _system_logger.log_event(
                level="info",
                module="nflx.conda",
                name="requested_packages",
                payload={
                    "full_id": newly_resolved_env_id.full_id,
                    "arch": newly_resolved_env_id.arch,
                    "req_id": newly_resolved_env_id.req_id,
                    "resolver_uuid": resolver.uuid,
                    "msg": str(newly_resolved_env.deps),
                },
            )

            # Log the sources
            _system_logger.log_event(
                level="info",
                module="nflx.conda",
                name="sources",
                payload={
                    "full_id": newly_resolved_env_id.full_id,
                    "arch": newly_resolved_env_id.arch,
                    "req_id": newly_resolved_env_id.req_id,
                    "resolver_uuid": resolver.uuid,
                    "msg": str(newly_resolved_env.sources),
                },
            )

            # Log the sys packages
            _system_logger.log_event(
                level="info",
                module="nflx.conda",
                name="sys_packages",
                payload={
                    "full_id": newly_resolved_env_id.full_id,
                    "arch": newly_resolved_env_id.arch,
                    "req_id": newly_resolved_env_id.req_id,
                    "resolver_uuid": resolver.uuid,
                    "msg": str(newly_resolved_env.extras),
                },
            )

            # Log the resolved packages
            _system_logger.log_event(
                level="info",
                module="nflx.conda",
                name="resolved_packages",
                payload={
                    "full_id": newly_resolved_env_id.full_id,
                    "arch": newly_resolved_env_id.arch,
                    "req_id": newly_resolved_env_id.req_id,
                    "resolver_uuid": resolver.uuid,
                    "msg": str(_format_packages(newly_resolved_env.packages)),
                },
            )



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Type

if TYPE_CHECKING:
    import metaflow_extensions.nflx.plugins.conda.conda
    import metaflow_extensions.nflx.plugins.conda.env_descr


from ..utils import CondaException


class Resolver:
    TYPES = ["invalid"]

    # Context:
    # https://sourcegraph.netflix.net/github.netflix.net/corp/mli-metaflow-custom/- \
    # /blob/nflx-metaflow/metaflow_extensions/nflx/plugins/conda \
    # /envsresolver.py?L480
    # "In PYPI-ONLY scenarios, we actually need a sub-environment to properly
    # resolve the pypi environment (due to https://github.com/pypa/pip/issues/11664).
    # We try to mutualize these environments."
    #
    # Originally we need only to check whether env_type == PYPI_ONLY to determine
    # whether a builder env is needed for a resolver.
    # (https://github.netflix.net/corp/mli-metaflow-custom/blob/ \
    # 4a2d72e25975d20f609c1c16f811b42805c504f1/nflx-metaflow/ \
    # metaflow_extensions/nflx/plugins/conda/envsresolver.py#L513)
    # As we introduce the uv managed environment (triggered by @uv decorator),
    # it's more convenient we define this flag REQUIRES_BUILDER_ENV to determine
    # whether a resolver requires a builder environment.
    REQUIRES_BUILDER_ENV = False

    _class_per_type = None  # type: Optional[Dict[str, Type[Resolver]]]

    @classmethod
    def _ensure_class_per_type(cls):
        if cls._class_per_type is None:
            cls._class_per_type = {t: c for c in cls.__subclasses__() for t in c.TYPES}

    @classmethod
    def get_resolver(cls, resolver_type: str):
        cls._ensure_class_per_type()
        assert cls._class_per_type

        # We will split uv and pip resolvers at some point but right now
        # we use uv where we can for the pip resolver so we currently consider "uv"
        # to be an alias of "pip" for now
        if resolver_type == "uv":
            resolver_type = "pip"
        resolver = cls._class_per_type.get(resolver_type)
        if resolver is None:
            raise CondaException(
                "Dependency resolver '%s' does not exist" % resolver_type
            )
        return resolver

    def __init__(self, conda: "metaflow_extensions.nflx.plugins.conda.conda.Conda"):
        self._conda = conda

    def resolve(
        self,
        env_type: "metaflow_extensions.nflx.plugins.conda.env_descr.EnvType",
        python_version_requested: str,
        deps: Dict[str, List[str]],
        sources: Dict[str, List[str]],
        extras: Dict[str, List[str]],
        architecture: str,
        builder_envs: Optional[
            List["metaflow_extensions.nflx.plugins.conda.env_descr.ResolvedEnvironment"]
        ] = None,
        base_env: Optional[
            "metaflow_extensions.nflx.plugins.conda.env_descr.ResolvedEnvironment"
        ] = None,
        file_paths: Dict[str, List[str]] = {},
    ) -> Tuple[
        "metaflow_extensions.nflx.plugins.conda.env_descr.ResolvedEnvironment",
        Optional[
            List["metaflow_extensions.nflx.plugins.conda.env_descr.ResolvedEnvironment"]
        ],
    ]:
        """
        Resolves the environment specified by the dependencies, the sources (channels
        or indices), extra information (used for Pypi resolvers) for the given
        architecture.

        The builder_envs are additional environments that may be used to help in
        the resolution.

        Parameters
        ----------
        deps : Dict[str, List[str]]
            Dependencies that need to be satisfied for this environment. The first key
            is the type of package (conda/pypi/etc)
        sources : Dict[str, List[str]]
            Sources to use (channels, indices, etc). The first key is the type of source
        extras : Dict[str, List[str]]
            Extra information (used by some resolver)
        architecture : str
            Architecture to resolve for
        builder_envs : Optional[List[ResolvedEnvironment]]
            Helper environments to use
        base_env : Optional[ResolvedEnvironment]
            Environment this environment is based off of (ie: it is being built on
            top of this environment).

        Returns
        -------
        Tuple[ResolvedEnvironment, Optional[List[ResolvedEnvironment]]]
            The first element of the tuple is this ResolvedEnvironment.
            The second element are the builder environments. This may
            return the same thing as the argument passed in or more environments if
            more were also built in the process.
        """
        raise NotImplementedError


# These need to be imported to "register"
from .builder_envs_resolver import BuilderEnvsResolver
from .conda_lock_resolver import CondaLockResolver
from .pip_resolver import PipResolver
from .pylock_toml_resolver import PylockTomlResolver



glue_script = """#! {python_executable}

import json
import os
import subprocess
import sys

import requests


def parse_args(args):
    # Poor man's parsing of arguments; we just need to extract channels and
    # whatever is at the end for the specs. We can ignore everything else
    channels = []
    virtual_package_channels = None
    specs = []
    next_is_channel = False
    after_channels = False
    for arg in args:
        if next_is_channel:
            channels.append(arg)
            virtual_package_channels = arg  # Virtual package one is always last
            next_is_channel = False
        elif arg == "--channel":
            next_is_channel = True
            after_channels = True
        elif after_channels:
            specs.append("==".join(arg.split(" ")))
    return channels, specs, virtual_package_channels


def extract_virtual_packages(path):
    virtual_packages = []
    with open(path, mode="r", encoding="utf-8") as f:
        virtual_package_descr = json.load(f)
    for pkg in virtual_package_descr["packages"].values():
        virtual_packages.append("=".join([pkg["name"], *pkg["version"].split("-")]))
    return virtual_packages


if __name__ == "__main__":
    platform = os.environ.get("CONDA_SUBDIR")
    if len(sys.argv) < 2 or sys.argv[1] != "create":
        # Pass to micromamba directly
        exit(
            subprocess.call(
                ["{micromamba_exec}", "-r", "{micromamba_root}"] + sys.argv[1:],
                shell=True
        ))
        
    channels, specs, virtual_package_channels = parse_args(sys.argv[1:])
    virtual_packages = extract_virtual_packages(
        os.path.join(virtual_package_channels[7:], platform, "repodata.json")
    )
    resp = requests.post(
        "http://localhost:{server_port}/solve",
        json= {{
            "specs": specs,
            "channels": channels,
            "platform": platform,
            "virtual_packages": virtual_packages,
        }},
    )
    if resp.status_code != 200:
        print(
            json.dumps(
                {{
                    "success": False,
                    "solver_problems": "Got %d status from micromamba server"
                    % resp.status_code,
                }}
            )
        )
        exit(1)
    json_response = resp.json()
    if "error_msg" in json_response:
        print(
            json.dumps(
                {{"success": False, "solver_problems": json_response["error_msg"]}}
            )
        )
        exit(1)
    # Here we create a response that is similar to what would be created by pure
    # micromamba but only need the FETCH portion as that is the only thing conda-lock
    # cares about in the end.
    print(
        json.dumps(
            {{
                "success": True,
                "dry_run": True,
                "prefix": "/dev/null",
                "actions": {{
                    "FETCH": json_response["packages"],
                    "LINK": [],
                    "PREFIX": "/dev/null",
                }},
            }}
        )
    )
"""



import json
import os
import re
import shutil
import subprocess
import sys
from typing import Any, Dict, Optional

from metaflow.extension_support import (
    dump_module_info,
    get_extensions_in_dir,
    update_package_info,
)

from metaflow.packaging_sys import MFCONTENT_MARKER, MetaflowCodeContentV1Base

# _deps_parse = re.compile(r"([^<>=!~]+)(.*)")
_ext_parse = re.compile(r"([-_\w]+)\(([^)]+)\)")
_git_version = re.compile(r"-git([0-9a-f]+)(-dirty)?$")

name_to_pkg = {"netflix-ext": "metaflow-netflixext"}

# NOT used for now -- used in the commented out list command from environment_cmd.py
# May require cleanup
# def parse_deps(deps: str) -> List[TStr]:
#     deps = deps.split(";")
#     result = []
#     for dep in deps:
#         match = _deps_parse.match(dep)
#         if match is None:
#             raise ValueError("Dependency %s is not a valid constraint")
#         pkg, version = match.groups()
#         pkg_with_category = pkg.split("::")
#         if len(pkg_with_category) > 1:
#             category = pkg_with_category[0]
#             pkg = pkg_with_category[1]
#         else:
#             category = "conda"
#         version = version.lstrip("=")
#         result.append(TStr(category, "%s==%s" % (pkg, version)))
#     return result


# def parse_channels(channels: Optional[str]) -> List[TStr]:
#     if not channels:
#         return []

#     result = []
#     channels = channels.split(",")
#     for c in channels:
#         channel_with_category = c.split("::")
#         if len(channel_with_category) > 1:
#             category = channel_with_category[0]
#             c = channel_with_category[1]
#         else:
#             category = "conda"
#         result.append(TStr(category, c))
#     return result

# def req_id_from_spec(python: str, deps: str, channels: str) -> str:
#    deps = parse_deps(deps)
#    channels = parse_channels(channels)
#    deps.append(TStr("conda", "python==%s" % python))
#    return ResolvedEnvironment.get_req_id(deps, channels)


# def local_instances_for_req_id(conda: Conda, req_id: str):
#     created_envs = conda.created_environments()  # This returns all MF environments
#     local_instances = {}  # type: Dict[str, List[str]]
#     for env_id, present_list in created_envs.items():
#         if env_id.req_id != req_id:
#             continue
#         local_instances[env_id.full_id] = present_list
#     return local_instances


def _merge_directories(src_dir, dest_dir):
    # Due to a bug in PIP, we can't use --target to install namespace packages
    # so we hack around it by merging directories manually.

    for root, dirs, files in os.walk(src_dir):
        # Determine the path of the current directory relative to src_dir
        relative_path = os.path.relpath(root, src_dir)
        # Determine the corresponding path in the destination directory
        dest_path = os.path.join(dest_dir, relative_path)

        # Create directories in the destination directory
        for dir_name in dirs:
            dest_dir_path = os.path.join(dest_path, dir_name)
            if not os.path.exists(dest_dir_path):
                os.makedirs(dest_dir_path)

        # Copy files to the destination directory
        for file_name in files:
            src_file_path = os.path.join(root, file_name)
            dest_file_path = os.path.join(dest_path, file_name)
            shutil.copy2(src_file_path, dest_file_path)


def download_mf_version(
    executable: str, version_str: str, extension_info: Dict[str, Any], echo, fail_hard
):
    dst_code_dir = MetaflowCodeContentV1Base._code_dir

    def echo_or_fail(msg):
        if fail_hard:
            raise RuntimeError(
                msg
                + ". Use --no-strict to install latest versions if possible instead."
            )
        echo("WARNING: " + msg + " -- installing latest version if possible.")

    def _install_pkg(pkg: str, ver: Optional[str]):
        try:
            os.makedirs(dst_code_dir, exist_ok=True)
            subprocess.check_call(
                [
                    executable,
                    "-m",
                    "pip",
                    "install",
                    "--quiet",
                    "-t",
                    dst_code_dir,
                    "--no-deps",
                    "%s==%s" % (pkg, ver) if ver else pkg,
                ]
            )
        except subprocess.CalledProcessError as e:
            raise RuntimeError(
                "Could not install version '%s' of '%s': %s" % (ver, pkg, e.stderr)
            ) from e

    if not version_str:
        raise ValueError("Unknown version of Metaflow")

    s = version_str.split("+", 1)
    if _git_version.search(s[0]):
        # This is not a "public" release so we install the latest
        echo_or_fail("Metaflow's version is non public (%s)" % s[0])
        _install_pkg("metaflow", None)
    else:
        _install_pkg("metaflow", s[0])
    if len(s) == 1:
        return
    # We now install the other packages (extensions).
    # If we have extension_info, we can get that information from there.
    # That is ideal if we have that. If not, we do our best from the version string
    # where packages are in the form name(vers);name(vers) but name is not necessarily
    # the package name.
    wrong_version_info = set()

    if extension_info:
        first_round = True
        for pkg_name, pkg_info in extension_info["installed"].items():
            if pkg_name.startswith("_pythonpath"):
                # Local package with *zero* information
                echo_or_fail("Unknown extension present at runtime")
                continue
            if pkg_info["package_version"] == "<unk>" or _git_version.search(
                pkg_info["package_version"]
            ):
                echo_or_fail(
                    "Extension '%s' has a non-public version (%s)"
                    % (pkg_info["extension_name"], pkg_info["package_version"])
                )
                pkg_version = pkg_info["dist_version"]
                wrong_version_info.add(pkg_name)
            else:
                pkg_version = pkg_info["package_version"]
            _install_pkg(pkg_name, pkg_version)
            if first_round:
                shutil.move(
                    os.path.join(dst_code_dir, "metaflow_extensions"),
                    os.path.join(dst_code_dir, "metaflow_extensions_tmp"),
                )
                first_round = False
            else:
                _merge_directories(
                    os.path.join(dst_code_dir, "metaflow_extensions"),
                    os.path.join(dst_code_dir, "metaflow_extensions_tmp"),
                )
                shutil.rmtree(os.path.join(dst_code_dir, "metaflow_extensions"))
    else:
        s = s[1].split(";")
        first_round = True
        for pkg_desc in s:
            m = _ext_parse.match(pkg_desc)
            if not m:
                # In some cases (older Metaflow), the version is not recorded so
                # we just install the latest
                echo_or_fail("Extension '%s' does not have a version" % pkg_desc)
                pkg_name, pkg_version = pkg_desc, None
            else:
                pkg_name, pkg_version = m.groups()
            pkg = name_to_pkg.get(pkg_name)
            if pkg is None:
                raise ValueError("Metaflow extension '%s' is not known" % pkg_name)
            _install_pkg(pkg, pkg_version)
            if first_round:
                shutil.move(
                    os.path.join(dst_code_dir, "metaflow_extensions"),
                    os.path.join(dst_code_dir, "metaflow_extensions_tmp"),
                )
                first_round = False
            else:
                _merge_directories(
                    os.path.join(dst_code_dir, "metaflow_extensions"),
                    os.path.join(dst_code_dir, "metaflow_extensions_tmp"),
                )
                shutil.rmtree(os.path.join(dst_code_dir, "metaflow_extensions"))
    # We now do a few things to make sure the Metaflow environment is recreated
    # as closely as possible:
    #  - add a __init__.py file to the metaflow_extensions directory to prevent
    #    other extensions from being loaded
    #  - create a INFO file with the extension information. This will allow for the
    #    __init__.py file (since otherwise it is an error) and will also remove
    #    conflicts when trying to load the extensions.
    #  - add the mf install file
    shutil.move(
        os.path.join(dst_code_dir, "metaflow_extensions_tmp"),
        os.path.join(dst_code_dir, "metaflow_extensions"),
    )
    sys.path.insert(0, dst_code_dir)
    installed_packages, pkgs_per_extension_point = get_extensions_in_dir(
        os.path.join(os.getcwd(), dst_code_dir)
    )
    # Update the information with the reported version and name from extension_info
    for pkg_name, pkg_info in extension_info["installed"].items():
        if pkg_name in installed_packages:
            update_package_info(
                pkg_to_update=installed_packages[pkg_name],
                dist_version=pkg_info["dist_version"],
                extension_name=pkg_info["extension_name"],
            )
            if pkg_name not in wrong_version_info:
                update_package_info(
                    pkg_to_update=installed_packages[pkg_name],
                    package_version=pkg_info["package_version"],
                )

    key, val = dump_module_info(installed_packages, pkgs_per_extension_point)
    sys.path.pop(0)

    with open(
        os.path.join(dst_code_dir, "metaflow_extensions", "__init__.py"),
        "w+",
        encoding="utf-8",
    ) as f:
        f.write("# This file is automatically generated by Metaflow\n")

    os.makedirs(MetaflowCodeContentV1Base._other_dir, exist_ok=True)
    with open(
        os.path.join(
            MetaflowCodeContentV1Base._other_dir, MetaflowCodeContentV1Base._info_file
        ),
        "w+",
        encoding="utf-8",
    ) as f:
        json.dump({key: val}, f)

    # This is attrocious -- see about fixing at some point. It's a small file though
    # and with local code packaging, this should not happen much anymore.
    with open(MFCONTENT_MARKER, mode="w", encoding="utf-8") as f:
        json.dump({"version": 1, "module_files": []}, f)



import os

from typing import Dict, List, Any, Optional
from metaflow import Task, namespace, Step, Run
from metaflow.plugins.env_escape import generate_trampolines


class CurrentStubGenerator(object):
    def __init__(self, task_pathspec):
        """
        Initializes the CurrentStubGenerator.

        Parameters
        ----------
        task_pathspec : str
            The pathspec of the task.
        """
        self._task_pathspec = task_pathspec
        self._task = Task(task_pathspec, _namespace_check=False)
        self._step_name = self.task.parent.id
        self._run_id = self.task.parent.parent.id
        self._flow_name = self.task.parent.parent.parent.id

    def get(self, key: str, default: Optional[Any] = None) -> Optional[Any]:
        """
        Returns the value of the attribute with the given key.

        Parameters
        ----------
        key : str
            The key of the attribute.
        default : Any, optional, default None
            The default value to return if the key is not found.

        Returns
        -------
        Any
            The value of the attribute with the given key, or the default value
            if the key is not found.
        """
        return getattr(self, key, default)

    def __getattr__(self, key: str, default: Optional[Any] = None) -> Optional[Any]:
        """
        Called when an attribute lookup has not found the attribute in the usual places.

        Parameters
        ----------
        key : str
            The key of the attribute.
        default : Any, optional, default None
            The default value to return if the key is not found.

        Raises
        ------
        AttributeError
            If the attribute does not exist.
        """
        raise AttributeError(
            "'{}' object has no attribute '{}'. The Current Stub does not support all the "
            "Current singleton attributes yet.".format(type(self).__name__, key)
        )

    def __contains__(self, key: str) -> bool:
        """
        Returns True if the attribute with the given key exists, False otherwise.

        Parameters
        ----------
        key : str
            The key of the attribute.

        Returns
        -------
        bool
            True if the attribute with the given key exists, False otherwise.
        """
        return getattr(self, key, None) is not None

    @property
    def is_running_flow(self) -> bool:
        """
        Always returns True as this is a stub generator for a running flow.

        Returns
        -------
        bool
            True
        """
        return True

    @property
    def flow_name(self) -> Optional[str]:
        """
        The name of the currently executing flow.

        Returns
        -------
        str, optional
            Flow name.
        """
        return self._flow_name

    @property
    def run_id(self) -> Optional[str]:
        """
        The run ID of the currently executing run.

        Returns
        -------
        str, optional
            Run ID.
        """
        return self._run_id

    @property
    def step_name(self) -> Optional[str]:
        """
        The name of the currently executing step.

        Returns
        -------
        str, optional
            Step name.
        """
        return self._step_name

    @property
    def task_id(self) -> Optional[str]:
        """
        The task ID of the currently executing task.

        Returns
        -------
        str, optional
            Task ID.
        """
        return self.task.id

    @property
    def retry_count(self) -> int:
        """
        The index of the task execution attempt.

        Returns
        -------
        int
            The index of the task execution attempt.
        """
        return int(self._task.metadata_dict.get("attempt", 0))

    @property
    def origin_run_id(self) -> Optional[str]:
        """
        The original run ID of the currently executing run.

        Returns
        -------
        str, optional
            Original run ID.
        """
        return str(self._task.metadata_dict.get("origin-run-id", None))

    @property
    def pathspec(self) -> str:
        """
        The pathspec of the currently executing task.

        Returns
        -------
        str
            Pathspec.
        """
        return self._task_pathspec

    @property
    def task(self) -> Task:
        """
        The currently executing task.

        Returns
        -------
        Task
            The currently executing task.
        """
        return self._task

    @property
    def run(self) -> Run:
        """
        The currently executing run.

        Returns
        -------
        Run
            The currently executing run.
        """
        return self.task.parent.parent



import os
import sys
import json
import shutil
import traceback
import subprocess
from metaflow import Task, S3
from typing import Any, Dict, Optional, Tuple
from metaflow.exception import CommandException
from metaflow.plugins.env_escape import generate_trampolines
from metaflow_extensions.netflix_ext.cmd.debug.constants import Constants


def get_code_package_location(task: Task) -> str:
    """
    Gets the code package location for the task.

    Parameters
    ----------
    task : Task
        The task object.

    Returns
    -------
    str
        The code package location for the task.
    """
    code_package = task.metadata_dict.get("code-package", "").strip()
    code_package = json.loads(code_package)
    return code_package.get("location")


def get_python_version(task: Task) -> str:
    """
    Gets the python version for the task.

    Parameters
    ----------
    task : Task
        The task object.

    Returns
    -------
    str
        The python version for the task.
    """
    python_version = task.metadata_dict.get(
        "python_version", Constants.DEFAULT_PYTHON_VERSION
    )
    return ".".join(python_version.split(".")[:2])


def copy_stub_generator_to_metaflow_root_dir(metaflow_root_dir: str):
    """
    Copies the debug and current stub generator to the metaflow root directory.

    Parameters
    ----------
    metaflow_root_dir : str
        The metaflow root directory.
    """
    debug_stub_generator_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "debug_stub_generator.py",
    )
    current_stub_generator_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "current_stub_generator.py",
    )
    shutil.copy(debug_stub_generator_path, metaflow_root_dir)
    shutil.copy(current_stub_generator_path, metaflow_root_dir)


def fetch_environment_type(task: Task) -> str:
    """
    Fetches the conda environment type for the task.

    Parameters
    ----------
    task : Task
        The task object.

    Returns
    -------
    str
        The conda environment type for the task.
    """
    conda_env_id = task.metadata_dict.get("conda_env_id", "").strip()
    if conda_env_id == "":
        return "default"
    elif conda_env_id.startswith("[") and conda_env_id.endswith("]"):
        return "new_conda"
    raise CommandException(
        "The conda environment type for the task is not supported. Please use the new conda decorators avaailable in "
        "Metaflow."
    )


def _set_python_environment(
    obj,
    task: Task,
    task_pathspec: str,
    override_env: Optional[str] = None,
    override_env_from_pathspec: Optional[str] = None,
) -> str:
    """
    Sets the python environment for the debug task.

    Parameters
    ----------
    obj : Any
        The command object.
    task : Task
        The task object.
    task_pathspec : str
        The pathspec of the task.
    override_env : Optional[str], optional, default None
        Name of the named environment to use for debugging the task. Overrides the environment used in the Task.
    override_env_from_pathspec : Optional[str], optional, default None
        Pathspec to use as environment for debugging the task. Overrides the environment used in the Task.

    Returns
    -------
    str
        The path to the python executable for the debug task.

    """
    if override_env is not None and override_env_from_pathspec is not None:
        raise CommandException(
            "Only one of --override-env and --override-env-from-pathspec can be specified"
        )

    conda_environment_type = fetch_environment_type(task)
    path_spec = "/".join(task_pathspec.split("/")[:-1])
    path_spec_formatted = path_spec.replace("/", "_").replace("-", "_")
    mf_python_path, mf_env, mf_env_name = sys.executable, None, None
    obj.echo(f"Conda environment type: {conda_environment_type}")

    if conda_environment_type == "new_conda":
        mf_env_name = override_env if override_env else path_spec_formatted.lower()
        path_spec = (
            override_env_from_pathspec if override_env_from_pathspec else path_spec
        )
        mf_env = [override_env] if override_env else ["--pathspec", f"{path_spec}"]
        obj.echo(
            f"Creating conda environment: {mf_env_name} with pathspec: {path_spec}"
        )
        conda_command = [
            "metaflow",
            "environment",
            "--quiet",
            "create",
            "--name",
            mf_env_name,
            "--install-notebook",
            "--force",
        ]
        conda_command.extend(mf_env)
        # Check if the pathspec is a conda environment
        try:
            result = subprocess.run(
                conda_command,
                check=True,
                capture_output=True,
            )
            # We have to parse the stderr and find the exact line where the python path is present
            for line in result.stderr.decode().split("\n"):
                if "/bin/python" in line:
                    mf_python_path = line
                    break
            return mf_python_path
        except subprocess.CalledProcessError as e:
            raise CommandException(
                f"Error creating conda environment: {e.stderr.decode()}"
            )
        except Exception as e:
            raise CommandException("Setting python environment failed: {}".format(e))
    return mf_python_path


def _extract_code_package(obj, task: Task, metaflow_root_dir: str):
    """
    Extracts the code package for the task in the metaflow root directory.

    Parameters
    ----------
    obj : Any
        The command object.
    task : Task
        The task object.
    metaflow_root_dir : str
        The metaflow root directory.
    """
    os.makedirs(metaflow_root_dir, exist_ok=True)
    code_package_path = os.path.join(metaflow_root_dir, "code_package.tar.gz")
    code_package_location = get_code_package_location(task)
    try:
        with S3() as s3:
            res = s3.get(code_package_location)
            with open(code_package_path, "wb") as f:
                f.write(res.blob)

        # We now untar the code package
        shutil.unpack_archive(code_package_path, metaflow_root_dir)
        obj.echo(f"Code package extracted to {metaflow_root_dir}")
    except Exception as e:
        raise CommandException(
            "Error in downloading or extracting code package: {}".format(e)
        )


def _find_kernel_name(
    python_executable: Optional[str] = None,
) -> Optional[Tuple[str, str]]:
    """
    Finds the jupyter kernel name for the python executable. This is a best effort
    function and may potentially fail in some cases. We return None in those cases
    and will not set the kernel in the generated notebook.

    Parameters
    ----------
    python_executable : Optional[str], optional, default None
        The python executable path.

    Returns
    -------
    Optional[Tuple[str, str, str]]
        The kernel name, kernel display name, and kernelspec path if found, None otherwise.
    """
    try:
        output = subprocess.check_output(["jupyter", "kernelspec", "list", "--json"])
        kernelspecs = json.loads(output)
        for kernel_name, kernel_spec in kernelspecs["kernelspecs"].items():
            if kernel_spec["spec"]["argv"][0] == python_executable:
                return (
                    kernel_name,
                    kernel_spec["spec"]["display_name"],
                    kernel_spec["resource_dir"],
                )
    except Exception as e:
        # Ignore the exception and return None as it is a best effort function
        print(f"Error finding kernel name: {traceback.format_exc()}")
        pass
    return None


def _generate_escape_trampolines(metaflow_root_dir: str):
    """
    Generates the escape trampolines for the task.
    """
    trampoline_dir = os.path.join(metaflow_root_dir, "_escape_trampolines")
    os.makedirs(trampoline_dir, exist_ok=True)
    generate_trampolines(trampoline_dir)



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false
import json
import os
import sys
import tempfile

from itertools import chain
from typing import Dict, List, Optional, Tuple, cast

# Consider re-adding if resolving fails weirdly.
# from metaflow.exception import retry_exp_backoff

from ..env_descr import (
    CondaPackageSpecification,
    PackageSpecification,  # noqa
    EnvType,
    ResolvedEnvironment,
)
from ..utils import (
    CondaException,
    channel_or_url,
    clean_up_double_equal,
    filter_user_reqs_by_markers,
    parse_explicit_url_conda,
    split_into_dict,
)
from . import Resolver


class CondaResolver(Resolver):
    TYPES = ["conda", "mamba", "micromamba"]

    # @retry_exp_backoff(
    #    (CondaException, OSError),
    #    "Failed to resolve conda environment",
    #    tries=3,
    #    max_delay_per_retry=5,
    # )
    def resolve(
        self,
        env_type: EnvType,
        python_version_requested: str,
        deps: Dict[str, List[str]],
        sources: Dict[str, List[str]],
        extras: Dict[str, List[str]],
        architecture: str,
        builder_envs: Optional[List[ResolvedEnvironment]] = None,
        base_env: Optional[ResolvedEnvironment] = None,
        file_paths: Dict[str, List[str]] = {},
        # full_id_unique_keys is not used in CondaResolver.
        full_id_unique_keys: Optional[Dict[str, str]] = None,
    ) -> Tuple[ResolvedEnvironment, Optional[List[ResolvedEnvironment]]]:
        if base_env:
            local_packages = [
                p for p in base_env.packages if not p.is_downloadable_url()
            ]
            if local_packages:
                raise CondaException(
                    "Local packages are not allowed in Conda: %s"
                    % ", ".join([p.package_name for p in local_packages])
                )
        deps = filter_user_reqs_by_markers(deps, python_version_requested, architecture)
        sys_overrides = split_into_dict(deps.get("sys", []))
        real_deps = clean_up_double_equal(
            chain(deps.get("conda", []), deps.get("npconda", []))
        )
        packages = []  # type: List[PackageSpecification]
        with tempfile.TemporaryDirectory() as mamba_dir:
            args = [
                "create",
                "--yes",
                "--prefix",
                os.path.join(mamba_dir, "prefix"),
                "--dry-run",
            ]
            have_channels = False
            for c in set(sources.get("conda", [])).difference(
                map(channel_or_url, self._conda.default_conda_channels)
            ):
                have_channels = True
                args.extend(["-c", c])

            if not have_channels:
                have_channels = any(["::" in d for d in real_deps])
            args.extend(real_deps)

            addl_env = {
                "CONDA_SUBDIR": architecture,
                "CONDA_PKGS_DIRS": os.path.join(mamba_dir, "pkgs"),
                "CONDA_ENVS_DIRS": os.path.join(mamba_dir, "envs"),
                "CONDA_ROOT": self._conda.root_prefix,
                "CONDA_UNSATISFIABLE_HINTS_CHECK_DEPTH": "0",
            }
            addl_env.update(
                {
                    "CONDA_OVERRIDE_%s"
                    % pkg_name[2:].upper(): pkg_version.split("=")[0]
                    for pkg_name, pkg_version in sys_overrides.items()
                }
            )
            if have_channels:
                # Add flexible-channel-priority because otherwise if a version
                # is present in the other channels but not in the higher
                # priority channels, it is not found.
                addl_env["CONDA_CHANNEL_PRIORITY"] = "flexible"
            conda_result = json.loads(self._conda.call_conda(args, addl_env=addl_env))

        # This returns a JSON blob with:
        #  - actions:
        #    - FETCH: List of objects to fetch -- this is where we get hash and URL
        #    - LINK: Packages to actually install (in that order)
        # On micromamba (or Mamba 2+), we can just use the LINK blob since it has all
        # information we need
        if not conda_result["success"]:
            print(
                "Pretty-printed Conda create result:\n%s" % conda_result,
                file=sys.stderr,
            )
            raise CondaException(
                "Could not resolve environment -- see above pretty-printed error."
            )

        if self._conda.is_non_conda_exec:
            for lnk in conda_result["actions"]["LINK"]:
                parse_result = parse_explicit_url_conda(
                    "%s#%s" % (lnk["url"], lnk["md5"])
                )
                packages.append(
                    CondaPackageSpecification(
                        filename=parse_result.filename,
                        url=parse_result.url,
                        url_format=parse_result.url_format,
                        hashes={parse_result.url_format: cast(str, parse_result.hash)},
                    )
                )
        else:

            def _pkg_key(
                name: str, platform: str, build_string: str, build_number: str
            ) -> str:
                return "%s_%s_%s_%s" % (name, platform, build_string, build_number)

            fetched_packages = {}  # type: Dict[str, Tuple[str, str]]
            for pkg in conda_result["actions"]["FETCH"]:
                fetched_packages[
                    _pkg_key(
                        pkg["name"], pkg["subdir"], pkg["build"], pkg["build_number"]
                    )
                ] = (pkg["url"], pkg["md5"])
            for lnk in conda_result["actions"]["LINK"]:
                k = _pkg_key(
                    lnk["name"],
                    lnk["platform"],
                    lnk["build_string"],
                    lnk["build_number"],
                )
                url, md5_hash = fetched_packages[k]
                if not url.startswith(lnk["base_url"]):
                    raise CondaException(
                        "Unexpected record for %s: %s" % (k, str(conda_result))
                    )
                parse_result = parse_explicit_url_conda("%s#%s" % (url, md5_hash))
                packages.append(
                    CondaPackageSpecification(
                        filename=parse_result.filename,
                        url=parse_result.url,
                        url_format=parse_result.url_format,
                        hashes={parse_result.url_format: cast(str, parse_result.hash)},
                    )
                )
        return (
            ResolvedEnvironment(
                deps,
                sources,
                extras,
                architecture,
                all_packages=packages,
                env_type=env_type,
                # full_id_unique_keys is a uv specific cache invalidation mechanism,
                # and PipResolver shouldn't need it. We still pass this along as {}
                # to make code consistent.
                full_id_unique_keys=full_id_unique_keys,
            ),
            builder_envs,
        )



import json
import os
import platform
import shutil
import subprocess
import tempfile
import time

from functools import wraps
from itertools import chain
from shutil import copy
from typing import Any, Dict, List, Optional, Set, Tuple, cast
from metaflow._vendor import click

from metaflow.cli import echo_dev_null, echo_always


from metaflow import Step, metaflow_config
from metaflow.exception import CommandException
from metaflow.package import MetaflowPackage
from metaflow.packaging_sys import MetaflowCodeContent
from metaflow.plugins import DATASTORES, METADATA_PROVIDERS
from metaflow.metaflow_config import (
    CONDA_ALL_ARCHS,
    CONDA_DEPENDENCY_RESOLVER,
    CONDA_TEST,
    DEFAULT_DATASTORE,
    DEFAULT_METADATA,
    get_pinned_conda_libs,
)
from metaflow_extensions.netflix_ext.plugins.conda.conda import (
    Conda,
    InvalidEnvironmentException,
)
from metaflow_extensions.netflix_ext.plugins.conda.env_descr import (
    EnvID,
    EnvType,
    ResolvedEnvironment,
    env_type_for_deps,
)
from metaflow_extensions.netflix_ext.plugins.conda.envsresolver import EnvsResolver

from metaflow_extensions.netflix_ext.plugins.conda.parsers import (
    parse_req_value,
    parse_toml_value,
    parse_yml_value,
)
from metaflow_extensions.netflix_ext.plugins.conda.utils import (
    AliasType,
    arch_id,
    channel_or_url,
    conda_deps_to_pypi_deps,
    get_sys_packages,
    resolve_env_alias,
    plural_marker,
    merge_dep_dicts,
    tstr_to_dict,
)

from metaflow._vendor.packaging.specifiers import SpecifierSet
from metaflow._vendor.packaging.utils import canonicalize_version

from .utils import download_mf_version


class CommandObj:
    def __init__(self):
        pass


def env_spec_options(func):
    @click.option(
        "--arch",
        default=None,
        multiple=True,
        help="Architecture to resolve for. Can be specified multiple times. "
        "If not specified, defaults to the architecture for the step or current machine "
        "if there is no step.",
        required=False,
    )
    @click.option(
        "--python",
        default=None,
        show_default=True,
        help="Python version for the environment -- defaults to current version",
    )
    @click.option(
        "-r",
        "--requirement",
        "req_file",
        default=None,
        type=click.Path(exists=True, readable=True, dir_okay=False, resolve_path=True),
        help="Get the environment definition from a requirements.txt file (pip format).",
    )
    @click.option(
        "-f",
        "--file",
        "yml_file",
        default=None,
        type=click.Path(exists=True, readable=True, dir_okay=False, resolve_path=True),
        help="Get the environment definition from a environment.yml file (conda-lock format).",
    )
    @click.option(
        "-p",
        "--pyproject",
        "pyproject_file",
        default=None,
        type=click.Path(exists=True, readable=True, dir_okay=False, resolve_path=True),
        help="Get the environment definition from a pyproject.toml file (poetry format).",
    )
    @click.option(
        "--using-pathspec",
        default=None,
        help="Environment created starting with the environment for the Pathspec specified",
    )
    @click.option(
        "--using",
        default=None,
        help="Environment created starting with the environment referenced here",
    )
    @click.option(
        "--from-pathspec",
        default=None,
        help="Environment created cloning the requirements in the Pathspec specified",
    )
    @click.option(
        "--from",
        "from_",
        default=None,
        help="Environment created cloning the requirements in the environment referenced here",
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@click.group()
@click.pass_context
def cli(ctx):
    pass


@cli.group(help="Execution environment related commands.")
@click.option(
    "--quiet/--no-quiet",
    show_default=True,
    default=False,
    help="Suppress unnecessary messages",
)
@click.option(
    "--metadata",
    default=DEFAULT_METADATA,
    show_default=True,
    type=click.Choice([m.TYPE for m in METADATA_PROVIDERS]),
    help="Metadata service type",
)
@click.option(
    "--datastore",
    default=DEFAULT_DATASTORE,
    show_default=True,
    type=click.Choice([d.TYPE for d in DATASTORES]),
    help="Datastore type",
)
# For later support of different environment types.
@click.option(
    "--environment",
    default="conda",
    show_default=True,
    type=click.Choice(["conda"]),
    help="Type of environment to manage",
)
@click.option(
    "--conda-root",
    default=None,
    help="Root path for Conda cached information. If not set, "
    "looks for METAFLOW_CONDA_S3ROOT (for S3)",
)
@click.option(
    "--quiet-file-output",
    default=None,
    hidden=True,
    type=str,
    help="Output the quiet output to this file; used for scripting",
)
@click.pass_context
def environment(
    ctx: Any,
    quiet: bool,
    metadata: str,
    datastore: str,
    environment: str,
    conda_root: Optional[str],
    quiet_file_output: Optional[str],
):
    if quiet:
        echo = echo_dev_null
    else:
        echo = echo_always

    obj = CommandObj()
    obj.quiet = quiet
    obj.echo = echo
    obj.echo_always = echo_always
    obj.datastore_type = datastore
    obj.quiet_file_output = quiet_file_output

    if conda_root:
        if obj.datastore_type == "s3":
            metaflow_config.CONDA_S3ROOT = conda_root
        elif obj.datastore_type == "azure":
            metaflow_config.CONDA_AZUREROOT = conda_root
        elif obj.datastore_type == "gs":
            metaflow_config.CONDA_GSROOT = conda_root
        else:
            # Should never happen due to click validation
            raise RuntimeError("Unknown datastore type: %s" % obj.datastore_type)
    obj.conda_root = getattr(
        metaflow_config, "CONDA_%sROOT" % obj.datastore_type.upper()
    )

    obj.metadata_type = metadata

    obj.conda = Conda(obj.echo, obj.datastore_type)
    ctx.obj = obj


@environment.command()
@click.option("--name", default=None, help="Name of the environment to create")
@click.option(
    "--local-only/--non-local",
    show_default=True,
    is_flag=True,
    default=False,
    help="Only create if environment is known locally",
)
@click.option(
    "--force/--no-force",
    default=False,
    is_flag=True,
    show_default=True,
    help="Recreate the environment if it already exists and remove the `into` directory "
    "if it exists",
)
@click.option(
    "--strict/--no-strict",
    default=True,
    is_flag=True,
    show_default=True,
    help="If True, fails if it cannot install the original Metaflow environment",
)
@click.option(
    "--into-dir",
    default=None,
    show_default=False,
    type=click.Path(file_okay=False, writable=True, readable=True, resolve_path=True),
    help="If the `envspec` refers to a Metaflow executed Step, downloads the step's "
    "code package into this directory",
)
@click.option(
    "--install-notebook/--no-install-notebook",
    default=False,
    show_default=True,
    help="Install the created environment as a Jupyter kernel -- requires --name.",
)
@click.option(
    "--pathspec",
    default=False,
    is_flag=True,
    show_default=True,
    help="The environment name given is a pathspec",
)
@click.argument("env-name")
@click.pass_obj
def create(
    obj,
    name: Optional[str],
    local_only: bool,
    force: bool,
    strict: bool,
    into_dir: Optional[str],
    install_notebook: bool,
    pathspec: bool,
    env_name: str,
):
    """
    Create a local environment based on env-name.

    env-name can either be:
      - a pathspec in the form <flowname>/<runid>/<stepname>
      - a partial hash (in the form of <req id>) -- this assumes the default environment
        for that hash.
      - a full hash (in the form of <req id>:<full id> or <full-id>
      - a name for the environment (as added using `--alias` when resolving the environment
        or using `metaflow environment alias`)
    """
    if into_dir:
        into_dir = os.path.abspath(into_dir)
        if os.path.exists(into_dir):
            if force:
                shutil.rmtree(into_dir)
                os.makedirs(into_dir)
            elif not os.path.isdir(into_dir) or len(os.listdir(into_dir)) != 0:
                raise CommandException(
                    "'%s' is not a directory or not empty -- use --force to force"
                    % into_dir
                )
        else:
            os.makedirs(into_dir)
    if install_notebook and name is None:
        raise click.BadOptionUsage("install-notebook", "requires --name")

    code_pkg = None
    mf_version = ""
    mf_extensions_info = None

    if pathspec:
        env_name = "step:%s" % env_name
    alias_type, resolved_alias = resolve_env_alias(env_name)
    if alias_type == AliasType.PATHSPEC:
        if not pathspec:
            raise click.BadOptionUsage("pathspec", "environment name is not a pathspec")
        step = Step(resolved_alias, _namespace_check=False)
        parameters_task = Step(
            "%s/_parameters" % step.parent.pathspec, _namespace_check=False
        ).task
        code_pkg = step.task.code
        # We use parameters_task to allow the creation of a environment for a task that
        # may not have fully run.
        mf_version = parameters_task.metadata_dict.get("metaflow_version", "")
        mf_extensions_info = parameters_task["_graph_info"].data.get("extensions")
    else:
        if pathspec:
            raise click.BadOptionUsage(
                "pathspec", "missing --pathspec; environment name is a pathspec"
            )

    env_id_for_alias = cast(Conda, obj.conda).env_id_from_alias(
        env_name, local_only=local_only
    )
    if env_id_for_alias is None:
        raise CommandException(
            "Environment '%s' does not refer to a known environment" % env_name
        )
    env = cast(Conda, obj.conda).environment(env_id_for_alias, local_only)
    if env is None:
        raise CommandException(
            "Environment '%s' does not exist for this architecture" % (env_name)
        )

    if install_notebook:
        start = time.time()
        # We need to install ipykernel into the resolved environment
        obj.echo("    Resolving an environment compatible with Jupyter ...", nl=False)

        # We first check if `ipykernel` already exists in the environment. If it does, we
        # can skip the whole resolution process.
        if not any("ipykernel" == p.package_name for p in env.packages):
            # We use envsresolver to properly deal with builder environments and what not
            resolver = EnvsResolver(obj.conda)
            # We force the env_type to be the same as the base env since we don't modify
            # that by adding these deps.

            # We also force the use of use_latest because we are not really doing
            # anything that would require a re-resolve (ie: the user doesn't really
            # care about the version of ipykernel most likely).
            resolver.add_environment(
                arch_id(),
                user_deps={
                    "pypi" if env.env_type == EnvType.PYPI_ONLY else "conda": [
                        "ipykernel"
                    ]
                },
                user_sources={},
                extras={},
                base_env=env,
                local_only=local_only,
                use_latest=":any:",
            )
            resolver.resolve_environments(obj.echo)
            update_envs = []  # type: List[ResolvedEnvironment]
            if obj.datastore_type != "local" or CONDA_TEST:
                # We may need to update caches
                # Note that it is possible that something we needed to resolve, we don't need
                # to cache (if we resolved to something already cached).
                formats = set()  # type: Set[str]
                for _, resolved_env, f, _ in resolver.need_caching_environments(
                    include_builder_envs=True
                ):
                    update_envs.append(resolved_env)
                    formats.update(f)

                cast(Conda, obj.conda).cache_environments(
                    update_envs, {"conda": list(formats)}
                )
            else:
                update_envs = [
                    resolved_env
                    for _, resolved_env, _ in resolver.new_environments(
                        include_builder_envs=True
                    )
                ]
            cast(Conda, obj.conda).add_environments(update_envs)

            # Update the default environment
            for _, resolved_env, _ in resolver.resolved_environments(
                include_builder_envs=True
            ):
                cast(Conda, obj.conda).set_default_environment(resolved_env.env_id)

            cast(Conda, obj.conda).write_out_environments()

            # We are going to be creating this new environment going forward (not the
            # initial env we got)
            _, env, _ = next(resolver.resolved_environments())

        delta_time = int(time.time() - start)
        obj.echo(" done in %d second%s." % (delta_time, plural_marker(delta_time)))

    name = name or "metaflowtmp_%s_%s" % (env.env_id.req_id, env.env_id.full_id)

    if obj.conda.created_environment(name):
        if not force:
            raise CommandException(
                "Environment '%s' already exists; use --force to force recreating"
                % name
            )
        obj.conda.remove_for_name(name)

    python_bin = None  # type: Optional[str]
    if into_dir:
        wrote_setup = False

        def _write_setup():
            env_dict = MetaflowCodeContent.get_env_vars_for_packaged_metaflow(
                os.getcwd()
            )
            if not env_dict:
                return False
            with open("metaflow_setup.sh", "w", encoding="utf-8") as f:
                f.write("#!/bin/sh\n")
                f.write(
                    "# This file is automatically generated by the "
                    "`metaflow environment create` command.\n"
                )
                f.write(
                    "# It sets up the environment for running Metaflow "
                    "in this directory.\n"
                )

                for k, v in env_dict.items():
                    if k.endswith(":"):
                        f.write("export %s=%s\n" % (k[:-1], v))
                    else:
                        f.write("export %s=%s\n" % (k, v))
                return True

        os.chdir(into_dir)
        python_bin = os.path.join(
            obj.conda.create_for_name(name, env, do_symlink=True), "bin", "python"
        )
        if code_pkg:
            start = time.time()
            obj.echo("    Downloading code package ...", nl=False)
            MetaflowPackage.cls_extract_into(
                code_pkg._code_metadata, code_pkg._code_obj, "."
            )
            code_pkg._code_obj.seek(0)
            delta_time = int(time.time() - start)
            obj.echo(" done in %d second%s." % (delta_time, plural_marker(delta_time)))

            # We write out the environment variables needed to use this environment
            # in a metaflow_setup.sh file.
            wrote_setup = _write_setup()
        elif alias_type == AliasType.PATHSPEC:
            # We get metaflow version and recreate it in the directory
            start = time.time()
            obj.echo(
                "Step '%s' does not have a code package -- "
                "downloading active Metaflow version only ..." % env_name,
                nl=False,
            )
            download_mf_version(
                "./__conda_python", mf_version, mf_extensions_info, obj.echo, strict
            )
            delta_time = int(time.time() - start)
            obj.echo(" done in %d second%s." % (delta_time, plural_marker(delta_time)))

            # We write out the environment variables needed to use this environment
            # in a metaflow_setup.sh file.
            wrote_setup = _write_setup()

        obj.echo(
            "Python executable `__conda_python` for environment '%s' downloaded "
            "into '%s'" % (env_name, into_dir)
        )
    else:
        python_bin = os.path.join(obj.conda.create_for_name(name, env), "bin", "python")

    if install_notebook:
        start = time.time()
        obj.echo("    Installing Jupyter kernel ...", nl=False)
        try:
            with tempfile.TemporaryDirectory() as d:
                kernel_info = {
                    "argv": [
                        python_bin,
                        "-m",
                        "ipykernel_launcher",
                        "-f",
                        "{connection_file}",
                    ],
                    "display_name": "Metaflow %s" % resolved_alias,
                    "language": "python",
                    "metadata": {"debugger": True},
                }
                if into_dir:
                    kernel_info["env"] = {"PYTHONPATH": into_dir}
                images_dir = os.path.abspath(
                    os.path.join(
                        os.path.dirname(os.path.abspath(__file__)),
                        "..",
                        "..",
                        "plugins",
                        "conda",
                        "resources",
                    )
                )
                for image in ("logo-32x32.png", "logo-64x64.png", "logo-svg.svg"):
                    copy(os.path.join(images_dir, image), os.path.join(d, image))
                with open(
                    os.path.join(d, "kernel.json"), mode="w", encoding="utf-8"
                ) as f:
                    json.dump(kernel_info, f)

                # Kernel name
                translation_table = str.maketrans(":/.-", "____")
                kernel_name = "mf_%s" % resolved_alias.translate(translation_table)
                try:
                    # Clean up in case it already exists -- ignore failures if it
                    # doesn't
                    subprocess.check_call(
                        ["jupyter", "kernelspec", "uninstall", "-y", kernel_name],
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )
                except subprocess.CalledProcessError:
                    pass
                subprocess.check_call(
                    ["jupyter", "kernelspec", "install", "--name", kernel_name, d],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
        except Exception as e:
            obj.echo("Error installing into Jupyter: %s" % e)
        else:
            delta_time = int(time.time() - start)
            obj.echo(" done in %d second%s." % (delta_time, plural_marker(delta_time)))

    if obj.quiet:
        obj.echo_always(python_bin)
        if obj.quiet_file_output:
            with open(obj.quiet_file_output, "w") as f:
                f.write(python_bin)
                f.write("\n")
    else:
        obj.echo(
            "Conda environment '%s' created locally, activate with "
            "`CONDA_ENVS_DIRS=%s %s activate %s`"
            % (
                name,
                obj.conda.root_env_dir,
                obj.conda._bins[CONDA_DEPENDENCY_RESOLVER],
                name,
            )
        )
        if into_dir and wrote_setup:
            obj.echo(
                f"Source `{into_dir}/metaflow_setup.sh` to use the installed "
                "Metaflow environment."
            )
    cast(Conda, obj.conda).write_out_environments()


@environment.command(help="Resolve an environment")
@click.option(
    "--alias",
    default=None,
    multiple=True,
    help="Alias the resolved environment; can be name or name:tag",
)
@click.option(
    "--force/--no-force",
    show_default=True,
    default=False,
    help="Force resolution of already resolved environments",
)
@click.option(
    "--local-only/--non-local",
    show_default=True,
    default=False,
    help="Only look locally for using environments",
)
@click.option(
    "--dry-run/--no-dry-run",
    show_default=True,
    default=False,
    help="Dry-run -- only resolve, do not download, cache, persist or alias anything",
)
@click.option(
    "--set-default/--no-set-default",
    show_default=True,
    default=True,
    help="Set the resolved environment as the default environment for this set of requirements",
)
@env_spec_options
@click.pass_obj
def resolve(
    obj: Any,
    alias: Optional[Tuple[str]],
    force: bool,
    local_only: bool,
    dry_run: bool,
    set_default: bool,
    arch: Optional[Tuple[str]],
    python: Optional[str],
    req_file: Optional[str],
    yml_file: Optional[str],
    pyproject_file: Optional[str],
    using_pathspec: Optional[str],
    using: Optional[str],
    from_pathspec: Optional[str],
    from_: Optional[str],
):
    # Check combinations -- TODO: we could leverage a click add-on to do this
    if req_file is not None and yml_file is not None and pyproject_file is not None:
        raise click.BadOptionUsage("-r/-f/-p", "Can only specify one of -r or -f or -p")
    if (
        using_pathspec is not None
        or using is not None
        or from_pathspec is not None
        or from_ is not None
    ) and python is not None:
        raise click.BadOptionUsage(
            "--python",
            "Cannot specify a Python version if using --using-pathspec/--using or --from-pathspec/--from",
        )
    if using_pathspec is not None and using is not None:
        raise click.BadOptionUsage(
            "--using-pathspec", "Cannot specify both --using-pathspec and --using"
        )
    if from_pathspec is not None and from_ is not None:
        raise click.BadOptionUsage(
            "--from-pathspec", "Cannot specify both --from-pathspec and --from"
        )

    using_str = None  # type: Optional[str]
    if using:
        using_str = using
    elif using_pathspec:
        using_str = "step:%s" % using_pathspec

    from_str = None  # type: Optional[str]
    if from_:
        from_str = from_
    elif from_pathspec:
        from_str = "step:%s" % from_pathspec

    if from_str and using_str:
        raise click.BadOptionUsage(
            "--using/--using-pathspec",
            "Cannot specify both --using/--using-pathspec and --from/--from-pathspec",
        )

    if from_str and (req_file or yml_file or pyproject_file):
        raise click.BadOptionUsage(
            "--from/--from-pathspec",
            "Cannot specify --from/--from-pathspec with -r/-f/-p",
        )

    resolver = EnvsResolver(obj.conda)
    new_conda_deps = {}  # type: Dict[str, str]
    new_pypi_deps = {}  # type: Dict[str, str]
    new_sys_deps = {}  # type: Dict[str, str]
    new_np_conda_deps = {}  # type: Dict[str, str]
    new_sources = {}  # type: Dict[str, List[str]]
    new_extras = {}  # type: Dict[str, List[str]]

    archs = list(arch) if arch else [arch_id()]
    base_env_id = None
    base_env = None  # type: Optional[ResolvedEnvironment]
    base_env_python = python  # type: Optional[str]

    if from_str:
        # If we have an environment to clone, we get the dependencies. This is useful
        # to "replatform" an environment.
        for arch in CONDA_ALL_ARCHS:
            base_env_id = cast(Conda, obj.conda).env_id_from_alias(
                from_str, arch, local_only
            )
            if base_env_id is None:
                continue
            base_env = cast(Conda, obj.conda).environment(
                EnvID(
                    req_id=base_env_id.req_id,
                    full_id=base_env_id.full_id,
                    arch=arch,
                ),
                local_only,
            )
            break
        if base_env is None:
            raise CommandException("No known environment for '%s'" % from_str)
        for p in base_env.packages:
            if p.package_name == "python":
                base_env_python = p.package_version
                break
        if base_env_python is None:
            raise InvalidEnvironmentException(
                "Cannot determine python version of from-pathspec environment"
            )

    if using_str:
        # We can pick any of the architecture to get the info about the base environment
        base_env_id = cast(Conda, obj.conda).env_id_from_alias(
            using_str, archs[0], local_only
        )
        if base_env_id is None:
            raise CommandException("No known environment for '%s'" % using_str)

        base_env = cast(Conda, obj.conda).environment(
            EnvID(
                req_id=base_env_id.req_id,
                full_id=base_env_id.full_id,
                arch=archs[0],
            ),
            local_only,
        )
        if base_env is None:
            raise CommandException(
                "Environment for '%s' is not available on architecture '%s'"
                % (using_str, archs[0])
            )

        for p in base_env.packages:
            if p.package_name == "python":
                base_env_python = p.package_version
                break
        if base_env_python is None:
            raise InvalidEnvironmentException(
                "Cannot determine python version of base environment"
            )

    # Parse yaml first to put conda sources first to be consistent with step decorator
    parsed_python_version = None
    if yml_file:
        parsed_python_version = _parse_yml_file(
            yml_file,
            new_extras,
            new_sources,
            new_conda_deps,
            new_pypi_deps,
            new_sys_deps,
        )
    if req_file:
        parsed_python_version = _parse_req_file(
            req_file,
            new_extras,
            new_sources,
            new_pypi_deps,
            new_np_conda_deps,
            new_sys_deps,
        )
    if pyproject_file:
        parsed_python_version = _parse_toml_file(
            pyproject_file,
            new_extras,
            new_sources,
            new_pypi_deps,
            new_np_conda_deps,
            new_sys_deps,
        )

    if base_env_python:
        if parsed_python_version:
            if from_str or using_str:
                # Check if the python version matches properly
                if not SpecifierSet(parsed_python_version).contains(base_env_python):
                    raise InvalidEnvironmentException(
                        "The base/from environment's Python version (%s) does not match the "
                        "one specified in the requirements file (%s)"
                        % (base_env_python, parsed_python_version)
                    )
            else:
                raise InvalidEnvironmentException(
                    "Cannot specify --python if the python dependency is already set "
                    "in the requirements file"
                )
    else:
        base_env_python = parsed_python_version
    if base_env_python is None:
        base_env_python = platform.python_version()

    # Compute the deps
    if from_str:
        deps = tstr_to_dict(cast(ResolvedEnvironment, base_env).deps)
    else:
        if len(new_conda_deps) == 0 and (
            not base_env or base_env.env_type == EnvType.PYPI_ONLY
        ):
            # Assume a pypi environment for base deps
            pypi_deps = conda_deps_to_pypi_deps(
                get_pinned_conda_libs(base_env_python, obj.datastore_type)
            )

            conda_deps = {}
        else:
            conda_deps = dict(
                get_pinned_conda_libs(base_env_python, obj.datastore_type)
            )
            pypi_deps = {}

        pypi_deps = merge_dep_dicts(pypi_deps, new_pypi_deps)
        conda_deps = merge_dep_dicts(conda_deps, new_conda_deps)

        deps = {
            "conda": [
                "%s==%s" % (name, ver) if ver else name
                for name, ver in conda_deps.items()
            ],
            "pypi": [
                "%s==%s" % (name, canonicalize_version(ver)) if ver else name
                for name, ver in pypi_deps.items()
            ],
            "npconda": [
                "%s==%s" % (name, ver) if ver else name
                for name, ver in new_np_conda_deps.items()
            ],
        }

    env_type = env_type_for_deps(deps)
    if not base_env:
        deps["conda"].append("python==%s" % base_env_python)
    else:
        # We try to see if we can keep the env_type as close as possible to base_env
        if base_env.env_type == EnvType.PYPI_ONLY and len(new_conda_deps) == 0:
            env_type = EnvType.PYPI_ONLY

    # Merge all sys deps so we have a union so we can get the same req ID for all
    # co-resolved envs
    sys_pkgs = {}
    for cur_arch in archs:
        sys_pkgs.update(
            get_sys_packages(cast(Conda, obj.conda).virtual_packages, cur_arch, False)
        )
    sys_deps = [
        "%s==%s" % (name, ver) if ver else name for name, ver in new_sys_deps.items()
    ]
    for p, v in sys_pkgs.items():
        if p not in new_sys_deps:
            sys_deps.append("%s==%s" % (p, v) if v else p)

    deps["sys"] = sys_deps

    for cur_arch in archs:
        if not from_str and base_env_id:
            # If we have have_from, we don't look for the environment for the arch
            # we are trying to resolve since it most likely doesn't exist. We keep
            # the base env we have and will use that as a basis to resolve for new
            # arch.
            base_env = cast(Conda, obj.conda).environment(
                EnvID(
                    req_id=base_env_id.req_id,
                    full_id=base_env_id.full_id,
                    arch=cur_arch,
                ),
                local_only,
            )
            if base_env is None:
                raise CommandException(
                    "Environment for '%s' is not available on architecture '%s'"
                    % (using_str, cur_arch)
                )

        # We add the default sources as well -- those sources go last and we convert
        # to simple channels if we can
        if from_str:
            sources = tstr_to_dict(cast(ResolvedEnvironment, base_env).sources)
        else:
            # Conda sources are always there since we get the base environment
            sources = {
                "conda": list(
                    dict.fromkeys(
                        map(
                            channel_or_url,
                            chain(
                                new_sources.get("conda", []),
                                obj.conda.default_conda_channels,
                            ),
                        )
                    )
                )
            }
            if env_type != EnvType.CONDA_ONLY:
                sources["pypi"] = list(
                    dict.fromkeys(
                        chain(
                            obj.conda.default_pypi_sources, new_sources.get("pypi", [])
                        )
                    )
                )

        resolver.add_environment(
            cur_arch,
            deps,
            sources,
            new_extras,
            base_env=base_env if using_str else None,
            base_from_full_id=(
                base_env.is_info_accurate if base_env and using_str else False
            ),
            local_only=local_only,
            force=force,
            force_co_resolve=len(archs) > 1,
        )

    has_something = False
    for _ in resolver.non_resolved_environments():
        has_something = True
        break
    if not has_something:
        resolved_env_id = next(resolver.all_environments())[1].env_id
        # Nothing to do
        if alias and not dry_run:
            # We don't care about arch for aliasing so pick one
            obj.echo(
                "No environments to resolve, aliasing only. Use --force to force "
                "re-resolution"
            )
            obj.conda.alias_environment(resolved_env_id, list(alias))
            cast(Conda, obj.conda).write_out_environments()
        else:
            raise CommandException(
                "No environments to resolve, use --force to force re-resolution"
            )
        return

    resolver.resolve_environments(obj.echo)

    resolved_env_id = next(resolver.all_environments())[1].env_id
    existing_envs = cast(Conda, obj.conda).created_environments(resolved_env_id.req_id)

    # Arch doesn't matter for aliasing
    if not dry_run and alias:
        obj.conda.alias_environment(resolved_env_id, list(alias))
    for env_id, env, _ in resolver.resolved_environments():
        if obj.quiet:
            obj.echo_always(env_id.arch)
            obj.echo_always(env.quiet_print(existing_envs.get(env.env_id)))
            if obj.quiet_file_output:
                with open(obj.quiet_file_output, "w") as f:
                    f.write(env_id.arch)
                    f.write("\n")
                    f.write(env.quiet_print(existing_envs.get(env.env_id)))
                    f.write("\n")
        else:
            obj.echo("### Environment for architecture %s" % env_id.arch)
            obj.echo(env.pretty_print(existing_envs.get(env.env_id)))

    if dry_run:
        obj.echo("Dry-run -- not caching or aliasing")
        return

    # If this is not a dry-run, we cache the environments and write out the resolved
    # information
    update_envs = []  # type: List[ResolvedEnvironment]
    if obj.datastore_type != "local" or CONDA_TEST:
        # We may need to update caches
        # Note that it is possible that something we needed to resolve, we don't need
        # to cache (if we resolved to something already cached).
        formats = set()  # type: Set[str]
        for _, resolved_env, f, _ in resolver.need_caching_environments(
            include_builder_envs=True
        ):
            update_envs.append(resolved_env)
            formats.update(f)

        cast(Conda, obj.conda).cache_environments(update_envs, {"conda": list(formats)})
    else:
        update_envs = [
            resolved_env
            for _, resolved_env, _ in resolver.new_environments(
                include_builder_envs=True
            )
        ]

    cast(Conda, obj.conda).add_environments(update_envs)

    # Update the default environment
    if set_default:
        for env_id, resolved_env, _ in resolver.all_environments(
            include_builder_envs=True
        ):
            obj.conda.set_default_environment(resolved_env.env_id)

    # We are done -- write back out the environments
    cast(Conda, obj.conda).write_out_environments()


@environment.command(help="Show information about an environment")
@click.option(
    "--local-only/--non-local",
    show_default=True,
    default=False,
    help="Only resolve source env using local information",
)
@click.option(
    "--arch",
    show_default=True,
    default=arch_id(),
    help="Show environment for this architecture",
)
@click.option(
    "--pathspec",
    default=False,
    is_flag=True,
    show_default=True,
    help="The environments given are pathspecs",
)
@click.argument("envs", required=True, nargs=-1)
@click.pass_obj
def show(obj, local_only: bool, arch: str, pathspec: bool, envs: Tuple[str]):
    # req-id -> full-id -> List of paths
    all_envs = {}  # Dict[str, Dict[str, List[str]]]
    created_envs = cast(Conda, obj.conda).created_environments()
    for env_id, present_list in created_envs.items():
        all_envs.setdefault(env_id.req_id, {})[env_id.full_id] = [
            os.path.basename(p) for p in present_list
        ]

    for env_name in envs:
        if pathspec:
            env_name = "step:%s" % env_name
        resolved_env = cast(Conda, obj.conda).environment_from_alias(
            env_name, arch, local_only
        )
        if resolved_env is None:
            if obj.quiet:
                obj.echo_always("%s@%s" % (env_name, arch))
                obj.echo_always("NOT_FOUND")
                if obj.quiet_file_output:
                    with open(obj.quiet_file_output, "w") as f:
                        f.write("%s@%s\n" % (env_name, arch))
                        f.write("NOT_FOUND\n")
            else:
                obj.echo(
                    "### Environment for '%s' (arch '%s') was not found"
                    % (env_name, arch)
                )
                obj.echo("")
        else:
            if obj.quiet:
                obj.echo_always("%s@%s" % (env_name, arch))
                obj.echo_always(
                    resolved_env.quiet_print(
                        all_envs.get(resolved_env.env_id.req_id, {}).get(
                            resolved_env.env_id.full_id
                        )
                    )
                )
                if obj.quiet_file_output:
                    with open(obj.quiet_file_output, "w") as f:
                        f.write("%s@%s\n" % (env_name, arch))
                        f.write(
                            resolved_env.quiet_print(
                                all_envs.get(resolved_env.env_id.req_id, {}).get(
                                    resolved_env.env_id.full_id
                                )
                            )
                        )
                        f.write("\n")
            else:
                obj.echo("### Environment for '%s' (arch '%s'):" % (env_name, arch))
                obj.echo(
                    resolved_env.pretty_print(
                        all_envs.get(resolved_env.env_id.req_id, {}).get(
                            resolved_env.env_id.full_id
                        )
                    )
                )


@environment.command(help="Alias an existing environment")
@click.option(
    "--local-only/--non-local",
    show_default=True,
    default=False,
    help="Only resolve source env using local information",
)
@click.option(
    "--pathspec",
    default=False,
    is_flag=True,
    show_default=True,
    help="The source environment given is a pathspec",
)
@click.argument("source_env")
@click.argument("alias")
@click.pass_obj
def alias(obj, local_only: bool, pathspec: bool, source_env: str, alias: str):
    if pathspec:
        source_env = "step:%s" % source_env
    env_id_for_alias = cast(Conda, obj.conda).env_id_from_alias(
        source_env, local_only=local_only
    )
    if env_id_for_alias is None:
        raise CommandException(
            "Environment '%s' does not refer to a known environment" % source_env
        )
    cast(Conda, obj.conda).alias_environment(env_id_for_alias, [alias])
    cast(Conda, obj.conda).write_out_environments()


@environment.command(
    help="Locally fetch an environment already resolved and present remotely"
)
@click.option(
    "--default/--no-default",
    default=True,
    show_default=True,
    help="Set the downloaded environment as default for its requirement ID",
)
@click.option(
    "--arch",
    default=None,
    show_default=True,
    help="Request this architecture -- defaults to current architecture if not specified",
)
@click.option(
    "--pathspec",
    default=False,
    is_flag=True,
    show_default=True,
    help="The environment name given is a pathspec",
)
@click.argument("source_env")
@click.pass_obj
def get(obj, default: bool, arch: Optional[str], pathspec: bool, source_env: str):
    if pathspec:
        source_env = "step:%s" % source_env
    env = cast(Conda, obj.conda).environment_from_alias(source_env, arch)
    if env is None:
        raise CommandException(
            "Environment '%s' does not refer to a known environment for %s"
            % (source_env, arch or arch_id())
        )
    if default:
        cast(Conda, obj.conda).set_default_environment(env.env_id)
    existing_envs_dict = cast(Conda, obj.conda).created_environments(env.env_id.req_id)
    if existing_envs_dict:
        existing_envs = existing_envs_dict.get(env.env_id, [])
    else:
        existing_envs = []
    if obj.quiet:
        obj.echo_always(env.quiet_print(existing_envs))
        if obj.quiet_file_output:
            with open(obj.quiet_file_output, "w") as f:
                f.write(env.quiet_print(existing_envs))
                f.write("\n")
    else:
        obj.echo(env.pretty_print(existing_envs))
    cast(Conda, obj.conda).write_out_environments()


def _parse_req_file(
    file_name: str,
    extra_args: Dict[str, List[str]],
    sources: Dict[str, List[str]],
    deps: Dict[str, str],
    np_deps: Dict[str, str],
    sys_deps: Dict[str, str],
) -> Optional[str]:
    with open(file_name, mode="r", encoding="utf-8") as f:
        return parse_req_value(f.read(), extra_args, sources, deps, np_deps, sys_deps)


def _parse_yml_file(
    file_name: str,
    _: Dict[str, List[str]],
    sources: Dict[str, List[str]],
    conda_deps: Dict[str, str],
    pypi_deps: Dict[str, str],
    sys_deps: Dict[str, str],
) -> Optional[str]:
    with open(file_name, mode="r", encoding="utf-8") as f:
        return parse_yml_value(f.read(), {}, sources, conda_deps, pypi_deps, sys_deps)


def _parse_toml_file(
    file_name: str,
    extra_args: Dict[str, List[str]],
    sources: Dict[str, List[str]],
    deps: Dict[str, str],
    np_deps: Dict[str, str],
    sys_deps: Dict[str, str],
) -> Optional[str]:
    with open(file_name, mode="r", encoding="utf-8") as f:
        return parse_toml_value(f.read(), extra_args, sources, deps, np_deps, sys_deps)


# @environment.command(help="List resolved environments for a set of dependencies")
# @click.option(
#     "--local-only/--no-local-only",
#     default=False,
#     show_default=True,
#     help="Only list environments locally known",
# )
# @click.option(
#     "--archs",
#     default=arch_id(),
#     show_default=True,
#     help="Comma separated list of architectures to list environments for",
# )
# @click.pass_obj
# @env_spec_options
# def list(obj, local_only, archs, python, deps, channels):
#     req_id = req_id_from_spec(python, deps, channels)
#     my_arch = arch_id()
#
#     obj.echo(
#         "Listing environments for python: %s, dependencies: %s and channels: %s "
#         "(requirement hash: %s)"
#         % (python, str(parse_deps(deps)), str(parse_channels(channels)), req_id)
#     )
#
#     # Get all the environments that we know about
#     envs = []
#     for arch in archs.split(","):
#         envs.extend(obj.conda.environments(req_id, arch, local_only))
#
#     # Get the local environments so we can say if an environment is present
#     local_instances = local_instances_for_req_id(obj.conda, req_id)
#
#     if obj.quiet:
#         obj.echo_always(
#             "# req_id full_id arch resolved_on resolved_by resolved_on packages local_instances"
#         )
#         for env_id, env in envs:
#             quiet_print_env(
#                 obj, env, local_instances=local_instances.get(env_id.full_id)
#             )
#     else:
#         # Extract the mapping between default and regular env
#         default_env_id = {}
#         for env_id, env in envs:
#             if env_id.full_id == "_default":
#                 default_env_id[env_id.arch] = env.env_id
#
#         # Print out the environments
#         for env_id, env in envs:
#             if env_id.full_id == "_default":
#                 # We will print it out as a regular environment
#                 continue
#             pretty_print_env(
#                 obj,
#                 env,
#                 env.env_id == default_env_id.get(env_id.arch),
#                 local_instances=local_instances.get(env_id.full_id),
#             )
#         if len(envs) == 0:
#             if local_only:
#                 obj.echo(
#                     "*No environments exist locally, try without "
#                     "--local-only to search remotely*"
#                 )
#             else:
#                 obj.echo("*No environments exist*")
#         if default_env_id is None:
#             obj.echo("\n\nNo default environment defined -- will resolve when needed")
